from urllib.parse import parse_qs, urlparse

import numpy
import pandas.testing
import pytest
from starlette.status import HTTP_400_BAD_REQUEST

from ..adapters.dataframe import DataFrameAdapter
from ..adapters.mapping import MapAdapter
from ..client import Context, from_context, record_history
from ..serialization.table import deserialize_arrow
from ..server.app import build_app
from .utils import URL_LIMITS, fail_with_status_code

tree = MapAdapter(
    {
        # a dataframe divided into three partitions
        "basic": DataFrameAdapter.from_pandas(
            pandas.DataFrame(
                {
                    "x": 1 * numpy.ones(10),
                    "y": 2 * numpy.ones(10),
                    "z": 3 * numpy.ones(10),
                }
            ),
            npartitions=3,
        ),
        # a dataframe with just one partition
        "single_partition": DataFrameAdapter.from_pandas(
            pandas.DataFrame(
                {
                    "x": 1 * numpy.ones(5),
                    "y": 2 * numpy.ones(5),
                    "z": 3 * numpy.ones(5),
                }
            ),
            npartitions=1,
        ),
        # a dataframe with many columns
        "wide": DataFrameAdapter.from_pandas(
            pandas.DataFrame({f"column_{i:03d}": i * numpy.ones(5) for i in range(10)}),
            npartitions=1,
        ),
        # a dataframe with mixed types
        "diverse": DataFrameAdapter.from_pandas(
            pandas.DataFrame(
                {
                    "A": numpy.array([1, 2, 3], dtype="|u8"),
                    "B": numpy.array([1, 2, 3], dtype="<f8"),
                    "C": ["one", "two", "three"],
                }
            ),
            npartitions=1,
        ),
    }
)


@pytest.fixture(scope="module")
def context():
    app = build_app(tree)
    with Context.from_app(app) as context:
        yield context


def test_dataframe_basic(context):
    client = from_context(context)
    expected = tree["basic"].read()
    actual = client["basic"].read()
    assert client["basic"].structure().npartitions == 3
    pandas.testing.assert_frame_equal(actual, expected)
    assert client["basic"].columns == list(expected.columns) == list(actual.columns)


def test_dataframe_column_access(context):
    client = from_context(context)
    expected_df = tree["basic"].read()
    for col in expected_df.columns:
        expected = expected_df[col].values
        actual = client["basic"][col]
        numpy.testing.assert_equal(actual, expected)


@pytest.mark.xfail(reason="HTTP API accepts only one column")
def test_dataframe_multicolumn_access(context):
    client = from_context(context)
    original_df = tree["basic"].read()
    columns = list(original_df.columns)
    assert len(columns) > 1
    client["basic"][columns]


def test_dataframe_multicolumn_read(context):
    client = from_context(context)
    original_df = tree["basic"].read()
    columns = list(original_df.columns)[::2]  # Select a subset of columns
    assert len(columns) > 1
    actual_df = client["basic"].read(columns)

    for col in columns:
        expected = original_df[col].values
        actual = actual_df[col]
        numpy.testing.assert_equal(actual, expected)


def test_dataframe_single_partition(context):
    client = from_context(context)
    expected = tree["single_partition"].read()
    actual = client["single_partition"].read()
    assert client["single_partition"].structure().npartitions == 1
    pandas.testing.assert_frame_equal(actual, expected)


def test_reading_diverse_dtypes(context):
    client = from_context(context)
    expected = tree["diverse"].read()
    actual = client["diverse"].read()
    pandas.testing.assert_frame_equal(actual, expected)

    for col in expected.columns:
        actual = client["diverse"][col].read()
        assert numpy.array_equal(expected[col], actual)


def test_dask(context):
    client = from_context(context, "dask")["basic"]
    expected = tree["basic"].read()
    pandas.testing.assert_frame_equal(client.read().compute(), expected)
    pandas.testing.assert_frame_equal(client.compute(), expected)


@pytest.mark.parametrize(
    "url_limit, expected_method",
    (
        (URL_LIMITS.HUGE, "GET"),  # URL query should fit in a GET request
        (URL_LIMITS.DEFAULT, None),  # Expected method is not specified
        (URL_LIMITS.TINY, "POST"),  # URL query is too long for a GET request
    ),
    indirect=["url_limit"],
)
def test_url_limit_bypass(context, url_limit, expected_method):
    "GET requests beyond the URL length limit should become POST requests."
    client = from_context(context)
    df_client = client["wide"]
    original_df = tree["wide"].read()
    columns = list(original_df.columns)[::2]  # Pick a subset of columns
    expected = tree["wide"].read(columns)
    assert list(expected.columns) == columns

    with record_history() as history:
        actual = df_client.read(columns)
        pandas.testing.assert_frame_equal(actual, expected)
        assert list(actual.columns) == columns

        requests = list(request for request in history.requests)
        assert len(requests) == df_client.structure().npartitions

        request_methods = list(request.method for request in requests)
        if expected_method == "POST":
            assert "POST" in request_methods  # At least one POST request
        elif expected_method == "GET":
            assert "POST" not in request_methods  # No POST request


@pytest.mark.parametrize("http_method", ("GET", "POST"))
@pytest.mark.parametrize("link", ("full", "partition"))
def test_http_fetch_columns(context, http_method, link):
    "GET requests beyond the URL length limit should become POST requests."
    if http_method not in ("GET", "POST"):
        pytest.fail(reason="HTTP method {http_method} is not expected.")

    client = from_context(context)
    url_path = client["wide"].item["links"][link]
    original_df = tree["wide"].read()
    columns = list(original_df.columns)[::2]  # Pick a subset of columns
    params = {
        **parse_qs(urlparse(url_path).query),
        "partition": 0,  # Used by /table/partition; ignored by /table/full
        "column": columns,
    }
    expected = tree["wide"].read(columns)
    assert list(expected.columns) == columns

    with record_history() as history:
        if http_method == "POST":
            body = params.pop("column")
            response = context.http_client.post(url_path, json=body, params=params)
        elif http_method == "GET":
            response = context.http_client.get(url_path, params=params)
        response.raise_for_status()
        actual = deserialize_arrow(response.read())
        pandas.testing.assert_frame_equal(actual, expected)
        assert list(actual.columns) == columns

        requests = list(request for request in history.requests)
        assert len(requests) == 1


def test_deprecated_query_parameter(context):
    "HTTP route /table/partition: 'field' is a deprecated query parameter"
    client = from_context(context)
    url_path = client["basic"].item["links"]["partition"]
    params = {
        **parse_qs(urlparse(url_path).query),
        "partition": 0,
        "field": "x",
    }
    with pytest.warns(DeprecationWarning, match=r"'field'"):
        context.http_client.get(url_path, params=params)


@pytest.mark.filterwarnings("ignore::DeprecationWarning")
def test_redundant_query_parameters(context):
    "HTTP route /table/partition accepts 'column' or 'field', but not both"
    client = from_context(context)
    url_path = client["basic"].item["links"]["partition"]
    original_params = {
        **parse_qs(urlparse(url_path).query),
        "partition": 0,
        "field": "x",
        "column": "y",
    }

    # It is OK to include query parameter 'column' OR 'field'
    for param in ("field", "column"):
        params = original_params.copy()
        params.pop(param)
        context.http_client.get(url_path, params=params).raise_for_status()

    # It is an error to include query parameter 'column' AND 'field'
    with fail_with_status_code(HTTP_400_BAD_REQUEST) as response:
        params = original_params
        context.http_client.get(url_path, params=params).raise_for_status()
        assert "'field'" in response.text
        assert "'column'" in response.text



from abc import abstractmethod
from collections.abc import Mapping
from typing import Any, Dict, List, Literal, Optional, Protocol, Set, Tuple, Union

import dask.dataframe
import pandas
import sparse
from numpy.typing import NDArray

from ..ndslice import NDSlice
from ..server.schemas import Principal
from ..storage import Storage
from ..structures.array import ArrayStructure
from ..structures.awkward import AwkwardStructure
from ..structures.core import Spec, StructureFamily
from ..structures.sparse import SparseStructure
from ..structures.table import TableStructure
from ..type_aliases import JSON, Filters, Scopes
from .awkward_directory_container import DirectoryContainer


class BaseAdapter(Protocol):
    supported_storage: Set[type[Storage]]

    # @abstractmethod
    # @classmethod
    # def from_catalog(
    #     cls, data_source: DataSource, node: Node, /, **kwargs: Optional[Any]
    # ) -> "BaseAdapter":
    #     pass

    @abstractmethod
    def metadata(self) -> JSON:
        pass

    @abstractmethod
    def specs(self) -> List[Spec]:
        pass


class ContainerAdapter(Mapping[str, "AnyAdapter"], BaseAdapter):
    structure_family: Literal[StructureFamily.container]

    @abstractmethod
    def structure(self) -> None:
        pass


class ArrayAdapter(BaseAdapter, Protocol):
    structure_family: Literal[StructureFamily.array]

    @abstractmethod
    def structure(self) -> ArrayStructure:
        pass

    @abstractmethod
    def read(self, slice: NDSlice) -> NDArray[Any]:
        pass

    # TODO Fix slice
    @abstractmethod
    def read_block(self, block: Tuple[int, ...]) -> NDArray[Any]:
        pass


class AwkwardAdapter(BaseAdapter, Protocol):
    structure_family: Literal[StructureFamily.awkward]

    @abstractmethod
    def structure(self) -> AwkwardStructure:
        pass

    @abstractmethod
    def read(self) -> NDArray[Any]:  # Are Slice and Array defined by numpy somewhere?
        pass

    @abstractmethod
    def read_buffers(self, form_keys: Optional[List[str]] = None) -> Dict[str, Any]:
        pass

    @abstractmethod
    def write(self, container: DirectoryContainer) -> None:
        pass


class SparseAdapter(BaseAdapter, Protocol):
    structure_family: Literal[StructureFamily.sparse] = StructureFamily.sparse

    @abstractmethod
    def structure(self) -> SparseStructure:
        pass

    # TODO Fix slice (just like array)
    def read(
        self, slice: NDSlice
    ) -> sparse.COO:  # Are Slice and Array defined by numpy somewhere?
        pass

    # TODO Fix slice (just like array)
    def read_block(self, block: Tuple[int, ...]) -> sparse.COO:
        pass


class TableAdapter(BaseAdapter, Protocol):
    structure_family: Literal[StructureFamily.table] = StructureFamily.table

    @abstractmethod
    def structure(self) -> TableStructure:
        pass

    @abstractmethod
    def read(
        self, fields: List[str]
    ) -> Union[dask.dataframe.DataFrame, pandas.DataFrame]:
        pass

    @abstractmethod
    def read_partition(
        self,
        partition: int,
        fields: Optional[str] = None,
    ) -> Union[dask.dataframe.DataFrame, pandas.DataFrame]:
        pass

    @abstractmethod
    def __getitem__(self, key: str) -> ArrayAdapter:
        pass


AnyAdapter = Union[
    ArrayAdapter, AwkwardAdapter, ContainerAdapter, SparseAdapter, TableAdapter
]


class AccessPolicy(Protocol):
    @abstractmethod
    async def allowed_scopes(
        self,
        node: BaseAdapter,
        principal: Principal,
        authn_access_tags: Optional[Set[str]],
        authn_scopes: Scopes,
    ) -> Scopes:
        pass

    @abstractmethod
    async def filters(
        self,
        node: BaseAdapter,
        principal: Principal,
        authn_access_tags: Optional[Set[str]],
        authn_scopes: Scopes,
        scopes: Scopes,
    ) -> Filters:
        pass



import cachetools

from ..adapters.resource_cache import default_resource_cache, with_resource_cache


def test_simple_cache():
    counter = 0

    def f():
        nonlocal counter
        counter += 1
        return "some value"

    cache = cachetools.Cache(maxsize=1)
    with_resource_cache("test_key", f, _resource_cache=cache)
    with_resource_cache("test_key", f, _resource_cache=cache)
    assert counter == 1


def test_default_cache():
    counter = 0

    def f():
        nonlocal counter
        counter += 1
        return "some value"

    cache = default_resource_cache()
    with_resource_cache("test_key", f, _resource_cache=cache)
    with_resource_cache("test_key", f, _resource_cache=cache)
    assert counter == 1


def test_cache_zero_size():
    counter = 0

    def f():
        nonlocal counter
        counter += 1
        return "some value"

    cache = cachetools.Cache(maxsize=0)
    with_resource_cache("test_key", f, _resource_cache=cache)
    with_resource_cache("test_key", f, _resource_cache=cache)
    assert counter == 2



"""Add structure_family to data_sources.

Revision ID: e756b9381c14
Revises: 2ca16566d692
Create Date: 2024-02-20 09:18:45.405078

"""
import sqlalchemy as sa
from alembic import op

from tiled.structures.core import StructureFamily

# revision identifiers, used by Alembic.
revision = "e756b9381c14"
down_revision = "2ca16566d692"
branch_labels = None
depends_on = None


def upgrade():
    connection = op.get_bind()
    # Initialize new structure_family column as a nullable at first because
    # existing rows will be empty.
    op.add_column(
        "data_sources",
        sa.Column("structure_family", sa.Enum(StructureFamily), nullable=True),
    )
    data_sources = sa.Table(
        "data_sources",
        sa.MetaData(),
        sa.Column("id", sa.Integer),
        sa.Column(
            "node_id",
            sa.Integer,
            sa.ForeignKey("nodes.id", ondelete="CASCADE"),
            nullable=False,
        ),
        sa.Column("structure_family", sa.Enum(StructureFamily), nullable=True),
    )
    # Fill in structure_family from associated node.
    results = connection.execute(
        sa.text(
            "SELECT data_sources.id, nodes.structure_family "
            "FROM data_sources JOIN nodes ON "
            "data_sources.node_id = nodes.id"
        )
    ).fetchall()
    for data_source_id, structure_family in results:
        connection.execute(
            data_sources.update()
            .values(structure_family=structure_family)
            .where(data_sources.c.id == data_source_id)
        )
    # Now make structure_family NOT NULL.
    if connection.engine.dialect.name == "sqlite":
        with op.batch_alter_table("data_sources") as batch_op:
            batch_op.alter_column("structure_family", nullable=False)
    else:
        op.alter_column("data_sources", "structure_family", nullable=False)


def downgrade():
    # This _could_ be implemented but we will wait for a need since we are
    # still in alpha releases.
    raise NotImplementedError



import copy
from pathlib import Path
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple, Union
from urllib.parse import quote_plus

import pandas
import pyarrow
import pyarrow.feather as feather
import pyarrow.fs

from ..catalog.orm import Node
from ..storage import FileStorage, Storage
from ..structures.core import Spec, StructureFamily
from ..structures.data_source import Asset, DataSource, Management
from ..structures.table import TableStructure
from ..type_aliases import JSON
from ..utils import ensure_uri, path_from_uri
from .array import ArrayAdapter
from .utils import init_adapter_from_catalog


class ArrowAdapter:
    """ArrowAdapter Class"""

    structure_family = StructureFamily.table
    supported_storage = {FileStorage}

    def __init__(
        self,
        data_uris: List[str],
        structure: Optional[TableStructure] = None,
        metadata: Optional[JSON] = None,
        specs: Optional[List[Spec]] = None,
        **kwargs: Optional[Any],
    ) -> None:
        """

        Parameters
        ----------
        data_uris : list of uris where data sits.
        structure :
        metadata :
        specs :
        """
        # TODO Store data_uris instead and generalize to non-file schemes.
        self._partition_paths = [path_from_uri(uri) for uri in data_uris]
        self._metadata = metadata or {}
        if structure is None:
            table = feather.read_table(self._partition_paths)
            structure = TableStructure.from_arrow_table(table)
        self._structure = structure
        self.specs = list(specs or [])

    @classmethod
    def from_catalog(
        cls,
        data_source: DataSource[TableStructure],
        node: Node,
        /,
        **kwargs: Optional[Any],
    ) -> "ArrowAdapter":
        return init_adapter_from_catalog(cls, data_source, node, **kwargs)  # type: ignore

    def metadata(self) -> JSON:
        return self._metadata

    @classmethod
    def init_storage(
        cls,
        storage: Storage,
        data_source: DataSource[TableStructure],
        path_parts: List[str],
    ) -> DataSource[TableStructure]:
        """
        Class to initialize the list of assets for given uri.
        Parameters
        ----------
        storage : the storage option for .arrow files
        data_source : data source representing the adapter
        path_parts: the list of partitions
        Returns
        -------
        The list of assets.
        """
        data_source = copy.deepcopy(data_source)  # Do not mutate caller input.
        data_uri = storage.uri + "".join(
            f"/{quote_plus(segment)}" for segment in path_parts
        )

        directory = path_from_uri(data_uri)
        directory.mkdir(parents=True, exist_ok=True)
        assets = [
            Asset(
                data_uri=f"{data_uri}/partition-{i}.arrow",
                is_directory=False,
                parameter="data_uris",
                num=i,
            )
            for i in range(data_source.structure.npartitions)
        ]
        data_source.assets.extend(assets)
        return data_source

    def structure(self) -> TableStructure:
        return self._structure

    def get(self, key: str) -> Union[ArrayAdapter, None]:
        if key not in self.structure().columns:
            return None
        return ArrayAdapter.from_array(self.read([key])[key].values)

    def generate_data_sources(
        self,
        mimetype: str,
        dict_or_none: Callable[[TableStructure], Dict[str, str]],
        item: Union[str, Path],
        is_directory: bool,
    ) -> List[DataSource[TableStructure]]:
        """

        Parameters
        ----------
        mimetype :
        dict_or_none :
        item :
        is_directory :

        Returns
        -------

        """
        return [
            DataSource(
                structure_family=self.structure_family,
                mimetype=mimetype,
                structure=self.structure(),
                parameters={},
                management=Management.external,
                assets=[
                    Asset(
                        data_uri=ensure_uri(item),
                        is_directory=is_directory,
                        parameter="data_uris",  # <-- PLURAL!
                        num=0,  # <-- denoting that the Adapter expects a list, and this is the first element
                    )
                ],
            )
        ]

    #
    @classmethod
    def from_single_file(
        cls,
        data_uri: str,
        structure: Optional[TableStructure] = None,
        metadata: Optional[JSON] = None,
        specs: Optional[List[Spec]] = None,
    ) -> "ArrowAdapter":
        """

        Parameters
        ----------
        data_uri :
        structure :
        metadata :
        specs :

        Returns
        -------

        """
        return cls(
            [data_uri],
            structure=structure,
            metadata=metadata,
            specs=specs,
        )

    def __getitem__(self, key: str) -> ArrayAdapter:
        """

        Parameters
        ----------
        key :

        Returns
        -------

        """
        # Must compute to determine shape.
        return ArrayAdapter.from_array(self.read([key])[key].values)

    def items(self) -> Iterator[Tuple[str, ArrayAdapter]]:
        yield from (
            (key, ArrayAdapter.from_array(self.read([key])[key].values))
            for key in self._structure.columns
        )

    def reader_handle_partiton(self, partition: int) -> pyarrow.RecordBatchFileReader:
        """
        Function to initialize and return the reader handle.
        Parameters
        ----------
        partition : the integer number corresponding to a specific partition.
        Returns
        -------
        The reader handle for specific partition.
        """
        if not Path(self._partition_paths[partition]).exists():
            raise ValueError(f"partition {partition} has not been stored yet")
        else:
            return pyarrow.ipc.open_file(self._partition_paths[partition])

    def reader_handle_all(self) -> Iterator[pyarrow.RecordBatchFileReader]:
        """
        Function to initialize and return the reader handle.
        Returns
        -------
        The reader handle.
        """
        for path in self._partition_paths:
            if not Path(path).exists():
                raise ValueError(f"path {path} has not been stored yet")
            else:
                with pyarrow.ipc.open_file(path) as reader:
                    yield reader

    def write_partition(
        self,
        data: Union[List[pyarrow.record_batch], pyarrow.record_batch, pandas.DataFrame],
        partition: int,
    ) -> None:
        """
        "Function to write the data into specific partition as arrow format."
        Parameters
        ----------
        data : data to write into arrow file. Can be a list of record batch, or pandas dataframe.
        partition: integer index of partition to be read.
        Returns
        -------
        """
        if isinstance(data, pandas.DataFrame):
            table = pyarrow.Table.from_pandas(data)
            batches = table.to_batches()
        else:
            if not isinstance(data, list):
                batches = [data]
            else:
                batches = data

        schema = batches[0].schema

        uri = self._partition_paths[partition]

        with pyarrow.ipc.new_file(uri, schema) as file_writer:
            for batch in batches:
                file_writer.write_batch(batch)

    def write(
        self,
        data: Union[List[pyarrow.record_batch], pyarrow.record_batch, pandas.DataFrame],
    ) -> None:
        """
        "Function to write the data as arrow format."
        Parameters
        ----------
        data : data to write into arrow file. Can be a list of record batch, or pandas dataframe.
        Returns
        -------
        """
        if isinstance(data, pandas.DataFrame):
            table = pyarrow.Table.from_pandas(data)
            batches = table.to_batches()
        else:
            if not isinstance(data, list):
                batches = [data]
            else:
                batches = data

        schema = batches[0].schema

        if self.structure().npartitions != 1:
            raise NotImplementedError
        uri = self._partition_paths[0]

        with pyarrow.ipc.new_file(uri, schema) as file_writer:
            for batch in data:
                file_writer.write_batch(batch)

    def read(self, fields: Optional[Union[str, List[str]]] = None) -> pandas.DataFrame:
        """
        The concatenated data from given set of partitions as pyarrow table.
        Parameters
        ----------
        Returns
        -------
        Returns the concatenated pyarrow table as pandas dataframe.
        """
        data = pyarrow.concat_tables(
            [partition.read_all() for partition in self.reader_handle_all()]
        )
        table = data.to_pandas()
        if fields is not None:
            return table[fields]
        return table

    def read_partition(
        self,
        partition: int,
        fields: Optional[Union[str, List[str]]] = None,
    ) -> pandas.DataFrame:
        """
        Function to read a batch of data from a given partition.
        Parameters
        ----------
        partition : the index of the partition to read.
        fields : optional fields parameter.

        Returns
        -------
        The pyarrow table corresponding to a given partition and batch as pandas dataframe.
        """
        reader = self.reader_handle_partiton(partition)
        table = reader.read_all().to_pandas()
        if fields is not None:
            return table[fields]
        return table



import math
import string
import warnings
from pathlib import Path

import dask.array
import httpx
import numpy
import pytest
from starlette.status import HTTP_400_BAD_REQUEST, HTTP_406_NOT_ACCEPTABLE

from ..adapters.array import ArrayAdapter
from ..adapters.mapping import MapAdapter
from ..client import Context, from_context
from ..serialization.array import as_buffer
from ..server.app import build_app
from .utils import fail_with_status_code

array_cases = {
    "b": (numpy.arange(10) % 2).astype("b"),
    "i": numpy.arange(-10, 10, dtype="i"),
    "uint8": numpy.arange(10, dtype="uint8"),
    "uint16": numpy.arange(10, dtype="uint16"),
    "uint64": numpy.arange(10, dtype="uint64"),
    "f": numpy.arange(10, dtype="f"),
    "c": (numpy.arange(10) * 1j).astype("c"),
    "m": numpy.array(["2007-07-13", "2006-01-13", "2010-08-13"], dtype="datetime64[D]")
    - numpy.datetime64("2008-01-01"),
    "M": numpy.array(["2007-07-13", "2006-01-13", "2010-08-13"], dtype="datetime64[D]"),
    "S": numpy.array([letter * 3 for letter in string.ascii_letters], dtype="S3"),
    "U": numpy.array([letter * 3 for letter in string.ascii_letters], dtype="U3"),
}
# TODO bitfield "t", void "v", and object "O" (which is not supported by default)
scalar_cases = {k: numpy.array(v[0], dtype=v.dtype) for k, v in array_cases.items()}
for v in scalar_cases.values():
    assert v.shape == ()
array_tree = MapAdapter({k: ArrayAdapter.from_array(v) for k, v in array_cases.items()})
scalar_tree = MapAdapter(
    {k: ArrayAdapter.from_array(v) for k, v in scalar_cases.items()}
)

cube_cases = {
    "tiny_cube": numpy.random.random((10, 10, 10)),
    "tiny_hypercube": numpy.random.random((10, 10, 10, 10, 10)),
}
cube_tree = MapAdapter({k: ArrayAdapter.from_array(v) for k, v in cube_cases.items()})
inf_tree = MapAdapter(
    {
        "example": ArrayAdapter.from_array(
            numpy.array([0, 1, numpy.nan, -numpy.inf, numpy.inf]),
            metadata={"infinity": math.inf, "-infinity": -math.inf, "nan": numpy.nan},
        )
    },
    metadata={"infinity": math.inf, "-infinity": -math.inf, "nan": numpy.nan},
)
arr_with_zero_dim = numpy.array([]).reshape((0, 100, 1, 10))
# Suppress RuntimeWarning: divide by zero encountered in true_divide
# from dask.array.core.
with warnings.catch_warnings():
    zero_tree = MapAdapter(
        {
            "example": ArrayAdapter.from_array(
                dask.array.from_array(arr_with_zero_dim, chunks=arr_with_zero_dim.shape)
            )
        }
    )

nd_array = numpy.arange(9).reshape((3, 3))
uniform_array = numpy.empty((3,), dtype=object)
for i in range(uniform_array.shape[0]):
    uniform_array[i] = nd_array[i]
ragged_array = numpy.array([numpy.arange(3), numpy.arange(4, 10)], dtype=object)
object_array = numpy.full((10,), {"a": 1}, dtype=object)
nested_arrays_tree = MapAdapter(
    {
        "uniform": ArrayAdapter.from_array(uniform_array),
        "ragged": ArrayAdapter.from_array(ragged_array),
        "objects": ArrayAdapter.from_array(object_array),
    }
)


@pytest.fixture(scope="module")
def context():
    tree = MapAdapter(
        {
            "array": array_tree,
            "cube": cube_tree,
            "inf": inf_tree,
            "scalar": scalar_tree,
            "zero": zero_tree,
            "nested_arrays": nested_arrays_tree,
        }
    )
    app = build_app(tree)
    with Context.from_app(app) as context:
        yield context


@pytest.mark.parametrize("kind", list(array_cases))
def test_array_dtypes(kind, context):
    client = from_context(context)["array"]
    expected = array_cases[kind]
    actual_via_slice = client[kind][:]
    actual_via_read = client[kind].read()
    assert numpy.array_equal(actual_via_slice, actual_via_read)
    assert numpy.array_equal(actual_via_slice, expected)


@pytest.mark.parametrize("kind", list(scalar_cases))
def test_scalar_dtypes(kind, context):
    client = from_context(context)["scalar"]
    expected = scalar_cases[kind]
    actual = client[kind].read()
    assert numpy.array_equal(actual, expected)


def test_shape_with_zero(context):
    client = from_context(context)["zero"]
    actual = client["example"].read()
    assert numpy.array_equal(actual, arr_with_zero_dim)


def test_nan_infinity_handler(tmpdir, context):
    client = from_context(context)["inf"]
    data = client["example"].read()
    assert numpy.isnan(data).any()
    assert numpy.isinf(data).any()
    metadata = tuple(client["example"].metadata.values())
    assert numpy.isnan(metadata).any()
    assert numpy.isinf(metadata).any()
    Path(tmpdir, "testjson").mkdir()
    client["example"].export(Path(tmpdir, "testjson", "test.json"))

    import json

    def strict_parse_constant(c):
        raise ValueError(f"{c} is not valid JSON")

    with open(Path(tmpdir, "testjson", "test.json"), "r") as json_file:
        open_json = json.load(json_file, parse_constant=strict_parse_constant)

    expected_list = [0.0, 1.0, None, None, None]
    assert open_json == expected_list


def test_block_validation(context):
    "Verify that block must be fully specified."
    client = from_context(context, "dask")["cube"]["tiny_cube"]
    block_url = httpx.URL(client.item["links"]["block"])
    # Malformed because it has only 2 dimensions, not 3.
    malformed_block_url = block_url.copy_with(params={"block": "0,0"})
    with fail_with_status_code(HTTP_400_BAD_REQUEST):
        client.context.http_client.get(malformed_block_url).raise_for_status()


def test_dask(context):
    expected = cube_cases["tiny_cube"]
    client = from_context(context, "dask")["cube"]["tiny_cube"]
    assert numpy.array_equal(client.read().compute(), expected)
    assert numpy.array_equal(client.compute(), expected)
    assert numpy.array_equal(client[:].compute(), expected)


def test_array_format_shape_from_cube(context):
    client = from_context(context)["cube"]
    with fail_with_status_code(HTTP_406_NOT_ACCEPTABLE):
        hyper_cube = client["tiny_hypercube"].export("test.png")  # noqa: F841


def test_array_interface(context):
    client = from_context(context)["array"]
    for k, v in client.items():
        assert v.shape == array_cases[k].shape
        assert v.ndim == array_cases[k].ndim
        assert v.nbytes == array_cases[k].nbytes
        assert v.dtype == array_cases[k].dtype
        assert numpy.array_equal(numpy.asarray(v), array_cases[k])
        # smoke test
        v.chunks
        v.dims


def test_uniform_nested_array_projected_to_ndarray(context):
    client = from_context(context)["nested_arrays"]["uniform"]
    assert client.dtype == numpy.int_
    assert client.read().dtype == numpy.int_
    assert numpy.array_equal(client.read(), nd_array)


@pytest.mark.parametrize("kind", ["ragged", "objects"])
def test_unparsable_nested_array_stringified(kind, context):
    # This behavior is due to the fact that ragged Numpy arrays, and those with
    # non-numeric types (except for strings) will likely have dtype=object,
    # which may not be parsable or reducible. As such we fallback to taking the
    # string representations of the array elements.
    client = from_context(context)["nested_arrays"][kind]
    assert "<U" in client.dtype.str
    assert "<U" in client.read().dtype.str
    assert isinstance(client[0], str)


@pytest.mark.parametrize("kind", list(array_cases))
def test_as_buffer(kind):
    output = as_buffer("application/octet-stream", array_cases[kind], {})
    assert len(output) == len(bytes(output))


@pytest.mark.parametrize(
    "chunks, expected",
    [
        ((3, 13, 17), "((3, 3, 3, 3, 3), (13,), (17,))"),
        ((1, 13, 17), "((1, 1, ..., 1), (13,), (17,))"),
        ((2, 13, 17), "((2, 2, ..., 2, 1), (13,), (17,))"),
        ((15, 13, 17), "((15,), (13,), (17,))"),
        (((1, 1, 1, 1, 1, 2, 2, 2, 2, 2), 13, 17), "(variable, (13,), (17,))"),
    ],
)
def test_array_client_repr(tmpdir, chunks, expected):
    arr = dask.array.random.random(size=(15, 13, 17), chunks=chunks)
    adapter = MapAdapter({"arr": ArrayAdapter.from_array(arr, dims=("x", "y", "z"))})
    app = build_app(adapter)
    with Context.from_app(app) as context:
        client = from_context(context)
        rep = repr(client["arr"])
        assert rep.startswith("<ArrayClient")
        assert "shape=(15, 13, 17)" in rep
        assert f"dtype={client['arr'].dtype}" in rep
        assert f"chunks={expected}" in rep
        if client["arr"].dims:
            assert "dims=('x', 'y', 'z')" in rep



import sys

import msgpack
import numpy as np
import pytest

from ..catalog import from_uri
from ..client import Context, from_context
from ..server.app import build_app

pytestmark = pytest.mark.skipif(
    sys.platform == "win32", reason="Requires Redis service"
)


@pytest.fixture
def tiled_websocket_context(tmpdir, redis_uri):
    """Fixture that provides a Tiled context with websocket support."""
    tree = from_uri(
        "sqlite:///:memory:",
        writable_storage=[
            f"file://localhost{str(tmpdir / 'data')}",
            f"duckdb:///{tmpdir / 'data.duckdb'}",
        ],
        readable_storage=None,
        init_if_not_exists=True,
        cache_settings={
            "uri": redis_uri,
            "ttl": 60,
            "socket_timeout": 10.0,
            "socket_connect_timeout": 10.0,
        },
    )

    app = build_app(
        tree,
        authentication={"single_user_api_key": "secret"},
    )

    with Context.from_app(app) as context:
        yield context


def test_subscribe_immediately_after_creation_websockets(tiled_websocket_context):
    context = tiled_websocket_context
    client = from_context(context)
    test_client = context.http_client

    # Create streaming array node using Tiled client
    arr = np.arange(10)
    streaming_node = client.write_array(arr, key="test_stream_immediate")

    # Connect WebSocket using TestClient with msgpack format and authorization
    with test_client.websocket_connect(
        "/api/v1/stream/single/test_stream_immediate?envelope_format=msgpack",
        headers={"Authorization": "Apikey secret"},
    ) as websocket:
        # Write updates using Tiled client
        for i in range(1, 4):
            new_arr = np.arange(10) + i
            streaming_node.write(new_arr)

        # Receive all updates
        received = []
        for _ in range(3):
            msg_bytes = websocket.receive_bytes()
            msg = msgpack.unpackb(msg_bytes)
            received.append(msg)

        # Verify all updates received in order
        assert len(received) == 3

        # Check that we received messages with the expected data
        for i, msg in enumerate(received):
            assert "timestamp" in msg
            assert "payload" in msg
            assert msg["shape"] == [10]

            # Verify payload contains the expected array data
            payload_array = np.frombuffer(msg["payload"], dtype=np.int64)
            expected_array = np.arange(10) + (i + 1)
            np.testing.assert_array_equal(payload_array, expected_array)


def test_websocket_connection_to_non_existent_node(tiled_websocket_context):
    """Test websocket connection to non-existent node returns 404."""
    context = tiled_websocket_context
    test_client = context.http_client

    non_existent_node_id = "definitely_non_existent_websocket_node_99999999"

    # Try to connect to websocket for non-existent node
    # This should result in an HTTP 404 response during the handshake
    response = test_client.get(
        f"/api/v1/stream/single/{non_existent_node_id}",
        headers={"Authorization": "Apikey secret"},
    )
    assert response.status_code == 404


def test_subscribe_after_first_update_websockets(tiled_websocket_context):
    """Client that subscribes after first update sees only subsequent updates."""
    context = tiled_websocket_context
    client = from_context(context)
    test_client = context.http_client

    # Create streaming array node using Tiled client
    arr = np.arange(10)
    streaming_node = client.write_array(arr, key="test_stream_after_update")

    # Write first update before subscribing
    first_update = np.arange(10) + 1
    streaming_node.write(first_update)

    # Connect WebSocket after first update
    with test_client.websocket_connect(
        "/api/v1/stream/single/test_stream_after_update?envelope_format=msgpack",
        headers={"Authorization": "Apikey secret"},
    ) as websocket:
        # Write more updates
        for i in range(2, 4):
            new_arr = np.arange(10) + i
            streaming_node.write(new_arr)

        # Should only receive the 2 new updates
        received = []
        for _ in range(2):
            msg_bytes = websocket.receive_bytes()
            msg = msgpack.unpackb(msg_bytes)
            received.append(msg)

        # Verify only new updates received
        assert len(received) == 2

        # Check that we received messages with the expected data
        for i, msg in enumerate(received):
            assert "timestamp" in msg
            assert "payload" in msg
            assert msg["shape"] == [10]

            # Verify payload contains the expected array data
            payload_array = np.frombuffer(msg["payload"], dtype=np.int64)
            expected_array = np.arange(10) + (
                i + 2
            )  # i+2 because we start from update 2
            np.testing.assert_array_equal(payload_array, expected_array)


def test_subscribe_after_first_update_from_beginning_websockets(
    tiled_websocket_context,
):
    """Client that subscribes after first update but requests from seq_num=0 sees all updates.

    Note: seq_num starts at 1 for the first data point. seq_num=0 means "start as far back
    as you have" (similar to Bluesky social)
    """
    context = tiled_websocket_context
    client = from_context(context)
    test_client = context.http_client

    # Create streaming array node using Tiled client
    arr = np.arange(10)
    streaming_node = client.write_array(arr, key="test_stream_from_beginning")

    # Write first update before subscribing
    first_update = np.arange(10) + 1
    streaming_node.write(first_update)

    # Connect WebSocket requesting from beginning
    with test_client.websocket_connect(
        "/api/v1/stream/single/test_stream_from_beginning?envelope_format=msgpack&start=0",
        headers={"Authorization": "Apikey secret"},
    ) as websocket:
        # First, should receive the initial array creation
        historical_msg_bytes = websocket.receive_bytes()
        historical_msg = msgpack.unpackb(historical_msg_bytes)
        assert "timestamp" in historical_msg
        assert "payload" in historical_msg
        assert historical_msg["shape"] == [10]

        # Verify historical payload (initial array creation - sequence 0)
        historical_payload = np.frombuffer(historical_msg["payload"], dtype=np.int64)
        expected_historical = np.arange(10)  # Initial array
        np.testing.assert_array_equal(historical_payload, expected_historical)

        # Next, should receive the first update (sequence 1)
        first_update_bytes = websocket.receive_bytes()
        first_update_msg = msgpack.unpackb(first_update_bytes)
        first_update_payload = np.frombuffer(
            first_update_msg["payload"], dtype=np.int64
        )
        expected_first_update = np.arange(10) + 1
        np.testing.assert_array_equal(first_update_payload, expected_first_update)

        # Write more updates
        for i in range(2, 4):
            new_arr = np.arange(10) + i
            streaming_node.write(new_arr)

        # Receive the new updates
        for i in range(2, 4):
            msg_bytes = websocket.receive_bytes()
            msg = msgpack.unpackb(msg_bytes)
            assert "timestamp" in msg
            assert "payload" in msg
            assert msg["shape"] == [10]

            # Verify payload contains the expected array data
            payload_array = np.frombuffer(msg["payload"], dtype=np.int64)
            expected_array = np.arange(10) + i
            np.testing.assert_array_equal(payload_array, expected_array)


def test_close_stream_success(tiled_websocket_context):
    """Test successful close of an existing stream."""
    context = tiled_websocket_context
    client = from_context(context)
    test_client = context.http_client

    # Create a streaming array node
    arr = np.arange(10)
    streaming_node = client.write_array(arr, key="test_close_stream")

    # Upload some data
    streaming_node.write(np.arange(10) + 1)

    # Add a small delay to ensure the stream is fully established
    import time

    time.sleep(0.5)

    # Now close the stream
    response = test_client.delete(
        "/api/v1/stream/close/test_close_stream",
        headers={"Authorization": "Apikey secret"},
    )
    assert response.status_code == 200

    # Now close the stream again
    response = test_client.delete(
        "/api/v1/stream/close/test_close_stream",
        headers={"Authorization": "Apikey secret"},
    )

    # TODO: I think the test is correct and the server should be updated.
    assert response.status_code == 404


def test_close_stream_not_found(tiled_websocket_context):
    """Test close endpoint returns 404 for non-existent node."""
    context = tiled_websocket_context
    test_client = context.http_client

    non_existent_node_id = "definitely_non_existent_node_99999999"

    response = test_client.delete(
        f"/api/v1/stream/close/{non_existent_node_id}",
        headers={"Authorization": "Apikey secret"},
    )
    assert response.status_code == 404


def test_websocket_connection_wrong_api_key(tiled_websocket_context):
    """Test websocket connection with wrong API key fails with 401."""
    from starlette.testclient import WebSocketDenialResponse

    context = tiled_websocket_context
    client = from_context(context)
    test_client = context.http_client

    # Create streaming array node using correct key
    arr = np.arange(10)
    client.write_array(arr, key="test_auth_websocket")

    # Try to connect to websocket with wrong API key
    with pytest.raises(WebSocketDenialResponse) as exc_info:
        with test_client.websocket_connect(
            "/api/v1/stream/single/test_auth_websocket?envelope_format=msgpack",
            headers={"Authorization": "Apikey wrong_key"},
        ):
            pass

    assert exc_info.value.status_code == 401


def test_close_stream_wrong_api_key(tiled_websocket_context):
    """Test close endpoint returns 403 with wrong API key."""
    context = tiled_websocket_context
    client = from_context(context)
    test_client = context.http_client

    # Create streaming array node using correct key
    arr = np.arange(10)
    client.write_array(arr, key="test_auth_close")

    # Try to close stream with wrong API key
    response = test_client.delete(
        "/api/v1/stream/close/test_auth_close",
        headers={"Authorization": "Apikey wrong_key"},
    )
    assert response.status_code == 401



import builtins
import copy
import os
from collections.abc import Mapping
from importlib.metadata import version
from typing import Any, Iterator, List, Optional, Tuple, Union, cast
from urllib.parse import quote_plus

import zarr.core
from numpy._typing import NDArray
from packaging.version import Version

from ..adapters.utils import IndexersMixin
from ..catalog.orm import Node
from ..iterviews import ItemsView, KeysView, ValuesView
from ..ndslice import NDSlice
from ..storage import FileStorage, Storage
from ..structures.array import ArrayStructure
from ..structures.core import Spec, StructureFamily
from ..structures.data_source import Asset, DataSource
from ..type_aliases import JSON
from ..utils import Conflicts, node_repr, path_from_uri
from .array import ArrayAdapter, slice_and_shape_from_block_and_chunks

ZARR_LIB_V2 = Version(version("zarr")) < Version("3")
if ZARR_LIB_V2:
    from zarr.storage import DirectoryStore as LocalStore
    from zarr.storage import init_array as create_array
else:
    from zarr import create_array
    from zarr.storage import LocalStore

INLINED_DEPTH = int(os.getenv("TILED_HDF5_INLINED_CONTENTS_MAX_DEPTH", "7"))


class ZarrArrayAdapter(ArrayAdapter):
    "Adapter for Zarr arrays"

    supported_storage = {FileStorage}

    @classmethod
    def init_storage(
        cls,
        storage: Storage,
        data_source: DataSource[ArrayStructure],
        path_parts: List[str],
    ) -> DataSource[ArrayStructure]:
        data_source = copy.deepcopy(data_source)  # Do not mutate caller input.
        data_uri = storage.uri + "".join(
            f"/{quote_plus(segment)}" for segment in path_parts
        )
        # Zarr requires evenly-sized chunks within each dimension.
        # Use the first chunk along each dimension.
        zarr_chunks = tuple(dim[0] for dim in data_source.structure.chunks)
        shape = tuple(dim[0] * len(dim) for dim in data_source.structure.chunks)
        directory = path_from_uri(data_uri)
        directory.mkdir(parents=True, exist_ok=True)
        store = LocalStore(str(directory))
        create_array(
            store,
            shape=shape,
            chunks=zarr_chunks,
            dtype=data_source.structure.data_type.to_numpy_dtype(),
        )
        data_source.assets.append(
            Asset(
                data_uri=data_uri,
                is_directory=True,
                parameter="data_uri",
            )
        )
        return data_source

    def _stencil(self) -> Tuple[slice, ...]:
        """Trim overflow because Zarr always has equal-sized chunks."""
        return tuple(builtins.slice(0, dim) for dim in self.structure().shape)

    def get(self, key: str) -> Union[ArrayAdapter, None]:
        return None

    def read(
        self,
        slice: NDSlice = NDSlice(...),
    ) -> NDArray[Any]:
        return self._array[self._stencil()][slice or ...]

    def read_block(
        self,
        block: Tuple[int, ...],
        slice: NDSlice = NDSlice(...),
    ) -> NDArray[Any]:
        block_slice, _ = slice_and_shape_from_block_and_chunks(
            block, self.structure().chunks
        )
        # Slice the block out of the whole array,
        # and optionally a sub-slice therein.
        return self._array[self._stencil()][block_slice][slice or ...]

    def write(
        self,
        data: NDArray[Any],
        slice: NDSlice = NDSlice(...),
    ) -> None:
        if slice:
            raise NotImplementedError
        self._array[self._stencil()] = data

    async def write_block(
        self,
        data: NDArray[Any],
        block: Tuple[int, ...],
    ) -> None:
        block_slice, shape = slice_and_shape_from_block_and_chunks(
            block, self.structure().chunks
        )
        self._array[block_slice] = data

    async def patch(
        self,
        data: NDArray[Any],
        offset: Tuple[int, ...],
        extend: bool = False,
    ) -> Tuple[Tuple[int, ...], Tuple[Tuple[int, ...], ...]]:
        """
        Write data into a slice of the array, maybe extending it.

        If the specified slice does not fit into the array, and extend=True, the
        array will be resized (expanded, never shrunk) to fit it.

        Parameters
        ----------
        data : array-like
        offset : tuple[int]
            Where to place the new data
        extend : bool
            If slice does not fit wholly within the shape of the existing array,
            reshape (expand) it to fit if this is True.

        Raises
        ------
        ValueError :
            If slice does not fit wholly with the shape of the existing array
            and expand is False
        """
        current_shape = self._array.shape
        normalized_offset = [0] * len(current_shape)
        normalized_offset[: len(offset)] = list(offset)
        new_shape = []
        slice_ = []
        for data_dim, offset_dim, current_dim in zip(
            data.shape, normalized_offset, current_shape
        ):
            new_shape.append(max(current_dim, data_dim + offset_dim))
            slice_.append(slice(offset_dim, offset_dim + data_dim))
        new_shape_tuple = tuple(new_shape)
        if new_shape_tuple != current_shape:
            if extend:
                # Resize the Zarr array to accommodate new data
                self._array.resize(new_shape_tuple)
            else:
                raise Conflicts(
                    f"Slice {slice} does not fit into array shape {current_shape}. "
                    "Use ?extend=true to extend array dimension to fit."
                )
        self._array[tuple(slice_)] = data
        new_chunks = []
        # Zarr has regularly-sized chunks, so no user input is required to
        # simply extend the existing pattern.
        for chunk_size, size in zip(self._array.chunks, new_shape_tuple):
            dim = [chunk_size] * (size // chunk_size)
            if size % chunk_size:
                dim.append(size % chunk_size)
            new_chunks.append(tuple(dim))
        new_chunks_tuple = tuple(new_chunks)
        return new_shape_tuple, new_chunks_tuple


class ZarrGroupAdapter(
    Mapping[str, Union["ArrayAdapter", "ZarrGroupAdapter"]],
    IndexersMixin,
):
    "Adapter for Zarr groups (containers)"

    structure_family = StructureFamily.container

    def __init__(
        self,
        zarr_group: zarr.Group,
        *,
        structure: Optional[ArrayStructure] = None,
        metadata: Optional[JSON] = None,
        specs: Optional[List[Spec]] = None,
    ) -> None:
        if structure is not None:
            raise ValueError(
                f"Structure is expected to be None for containers, not {structure}"
            )
        self._zarr_group = zarr_group
        self.specs = specs or []
        self._provided_metadata = metadata or {}
        super().__init__()

    def __repr__(self) -> str:
        return node_repr(self, list(self))

    def metadata(self) -> dict[str, Any]:
        return (
            {"attributes": self._zarr_group.attrs.asdict()}
            if ZARR_LIB_V2
            else cast(dict[str, Any], self._zarr_group.metadata.to_dict())
        )

    def structure(self) -> None:
        return None

    def __iter__(self) -> Iterator[Any]:
        yield from self._zarr_group

    def __getitem__(self, key: str) -> Union[ArrayAdapter, "ZarrGroupAdapter"]:
        value = self._zarr_group[key]
        if isinstance(value, zarr.Group):
            return ZarrGroupAdapter(value)
        else:
            return ZarrArrayAdapter.from_array(value)

    def __len__(self) -> int:
        return len(self._zarr_group)

    def keys(self) -> KeysView:  # type: ignore
        return KeysView(lambda: len(self), self._keys_slice)

    def values(self) -> ValuesView:  # type: ignore
        return ValuesView(lambda: len(self), self._items_slice)

    def items(self) -> ItemsView:  # type: ignore
        return ItemsView(lambda: len(self), self._items_slice)

    def search(self, query: Any) -> None:
        raise NotImplementedError

    def read(self, fields: Optional[str]) -> "ZarrGroupAdapter":
        if fields is not None:
            raise NotImplementedError
        return self

    # The following two methods are used by keys(), values(), items().

    def _keys_slice(
        self, start: int, stop: int, direction: int, page_size: Optional[int] = None
    ) -> List[Any]:
        keys = list(self._zarr_group)
        if direction < 0:
            keys = list(reversed(keys))
        return keys[start:stop]

    def _items_slice(
        self, start: int, stop: int, direction: int, page_size: Optional[int] = None
    ) -> List[Any]:
        items = [(key, self[key]) for key in list(self)]
        if direction < 0:
            items = list(reversed(items))
        return items[start:stop]

    def inlined_contents_enabled(self, depth: int) -> bool:
        return depth <= INLINED_DEPTH


class ZarrAdapter:
    @classmethod
    def from_catalog(
        cls,
        # A Zarr node may reference an array or a group (container).
        data_source: DataSource[Union[ArrayStructure, None]],
        node: Node,
        /,
        **kwargs: Optional[Any],
    ) -> Union[ZarrGroupAdapter, ArrayAdapter]:
        zarr_obj = zarr.open(
            path_from_uri(data_source.assets[0].data_uri)
        )  # Group or Array
        if node.structure_family == StructureFamily.container:
            return ZarrGroupAdapter(
                zarr_obj,
                structure=data_source.structure,
                metadata=node.metadata_,
                specs=node.specs,
                **kwargs,
            )
        else:
            return ZarrArrayAdapter(
                zarr_obj,
                structure=cast(ArrayStructure, data_source.structure),
                metadata=node.metadata_,
                specs=node.specs,
                **kwargs,
            )

    @classmethod
    def from_uris(
        cls, data_uri: str, **kwargs: Optional[Any]
    ) -> Union[ZarrArrayAdapter, ZarrGroupAdapter]:
        zarr_obj = zarr.open(path_from_uri(data_uri))  # Group or Array
        if isinstance(zarr_obj, zarr.Group):
            return ZarrGroupAdapter(zarr_obj, **kwargs)
        else:
            structure = ArrayStructure.from_array(zarr_obj)
            return ZarrArrayAdapter(zarr_obj, structure=structure, **kwargs)



from typing import Any, Dict, List, Optional, Set, Tuple, Union

import dask.dataframe
import numpy
import pandas
import sparse
from numpy._typing import NDArray

from ..ndslice import NDSlice
from ..storage import Storage
from ..structures.array import BuiltinDtype
from ..structures.core import Spec, StructureFamily
from ..structures.sparse import COOStructure
from ..type_aliases import JSON
from .array import slice_and_shape_from_block_and_chunks


class COOAdapter:
    "Wrap sparse Coordinate List (COO) arrays."
    structure_family = StructureFamily.sparse
    supported_storage: Set[type[Storage]] = set()

    @classmethod
    def from_arrays(
        cls,
        coords: NDArray[Any],
        data: Union[dask.dataframe.DataFrame, pandas.DataFrame],
        shape: Tuple[int, ...],
        dims: Optional[Tuple[str, ...]] = None,
        metadata: Optional[JSON] = None,
        specs: Optional[List[Spec]] = None,
    ) -> "COOAdapter":
        """
        Simplest constructor. Single chunk from coords, data arrays.

        Parameters
        ----------
        coords :
        data :
        shape :
        dims :
        metadata :
        specs :

        Returns
        -------

        """
        structure = COOStructure(
            dims=dims,
            shape=shape,
            chunks=tuple((dim,) for dim in shape),
            data_type=BuiltinDtype.from_numpy_dtype(data.dtype),
            coord_data_type=BuiltinDtype.from_numpy_dtype(coords.dtype),
            resizable=False,
        )
        return cls(
            {(0, 0): (coords, data)},
            structure,
            metadata=metadata,
            specs=specs,
        )

    @classmethod
    def from_coo(
        cls,
        coo: sparse.COO,
        *,
        dims: Optional[Tuple[str, ...]] = None,
        metadata: Optional[JSON] = None,
        specs: Optional[List[Spec]] = None,
    ) -> "COOAdapter":
        """
        Construct from sparse.COO object.
        Parameters
        ----------
        coo :
        dims :
        metadata :
        specs :

        Returns
        -------

        """
        return cls.from_arrays(
            coords=coo.coords,
            data=coo.data,
            shape=coo.shape,
            dims=dims,
            metadata=metadata,
            specs=specs,
        )

    @classmethod
    def from_global_ref(
        cls,
        blocks: Dict[Tuple[int, ...], Tuple[NDArray[Any], Any]],
        shape: Tuple[int, ...],
        chunks: Tuple[Tuple[int, ...], ...],
        *,
        dims: Optional[Tuple[str, ...]] = None,
        metadata: Optional[JSON] = None,
        specs: Optional[List[Spec]] = None,
    ) -> "COOAdapter":
        """
        Construct from blocks with coords given in global reference frame.
        Parameters
        ----------
        blocks :
        shape :
        chunks :
        dims :
        metadata :
        specs :

        Returns
        -------

        """
        local_blocks = {}
        for block, (coords, data) in blocks.items():
            offsets = []
            for b, c in zip(block, chunks):
                offset = sum(c[:b])
                offsets.append(offset)
            local_coords = coords - [[i] for i in offsets]
            local_blocks[block] = local_coords, data
        structure = COOStructure(
            dims=dims,
            shape=shape,
            chunks=chunks,
            data_type=BuiltinDtype.from_numpy_dtype(data.dtype),
            coord_data_type=BuiltinDtype.from_numpy_dtype(coords.dtype),
            resizable=False,
        )
        return cls(
            local_blocks,
            structure,
            metadata=metadata,
            specs=specs,
        )

    def __init__(
        self,
        blocks: Dict[Tuple[int, ...], Tuple[NDArray[Any], Any]],
        structure: COOStructure,
        *,
        metadata: Optional[JSON] = None,
        specs: Optional[List[Spec]] = None,
    ) -> None:
        """
        Construct from blocks with coords given in block-local reference frame.
        Parameters
        ----------
        blocks :
        structure :
        metadata :
        specs :
        """
        self.blocks = blocks
        self._metadata = metadata or {}
        self._structure = structure
        self.specs = specs or []

    def metadata(self) -> JSON:
        """

        Returns
        -------

        """
        return self._metadata

    def structure(self) -> COOStructure:
        """

        Returns
        -------

        """
        return self._structure

    def read_block(
        self, block: Tuple[int, ...], slice: NDSlice = NDSlice(...)
    ) -> sparse.COO:
        """

        Parameters
        ----------
        block :
        slice :

        Returns
        -------

        """
        coords, data = self.blocks[block]
        _, shape = slice_and_shape_from_block_and_chunks(block, self._structure.chunks)
        arr = sparse.COO(data=data[:], coords=coords[:], shape=shape)
        return arr[slice] if slice else arr

    def read(self, slice: NDSlice = NDSlice(...)) -> sparse.COO:
        """

        Parameters
        ----------
        slice :

        Returns
        -------

        """
        all_coords = []
        all_data = []
        for block, (coords, data) in self.blocks.items():
            offsets = []
            for b, c in zip(block, self._structure.chunks):
                offset = sum(c[:b])
                offsets.append(offset)
            global_coords = coords + [[i] for i in offsets]
            all_coords.append(global_coords)
            all_data.append(data)
        arr = sparse.COO(
            data=numpy.concatenate(all_data),
            coords=numpy.concatenate(all_coords, axis=-1),
            shape=self._structure.shape,
        )
        return arr[slice] if slice else arr



from pathlib import Path
from typing import Any, Callable, Generator, Union, cast

import numpy as np
import pyarrow as pa
import pytest

from tiled.adapters.sql import (
    COLUMN_NAME_PATTERN,
    TABLE_NAME_PATTERN,
    SQLAdapter,
    is_safe_identifier,
)
from tiled.storage import SQLStorage, get_storage, parse_storage, register_storage
from tiled.structures.core import StructureFamily
from tiled.structures.data_source import DataSource, Management
from tiled.structures.table import TableStructure
from tiled.utils import sanitize_uri

names = ["f0", "f1", "f2", "f3"]
data0 = [
    pa.array([1, 2, 3, 4, 5]),
    pa.array([1.0, 2.0, 3.0, 4.0, 5.0]),
    pa.array(["foo0", "bar0", "baz0", None, "goo0"]),
    pa.array([True, None, False, True, None]),
]
data1 = [
    pa.array([6, 7, 8, 9, 10, 11, 12]),
    pa.array([6.0, 7.0, 8.0, 9.0, 10.0, 11.0, 12.0]),
    pa.array(["foo1", "bar1", None, "baz1", "biz", None, "goo"]),
    pa.array([None, True, True, False, False, None, True]),
]
data2 = [
    pa.array([13, 14]),
    pa.array([13.0, 14.0]),
    pa.array(["foo2", "baz2"]),
    pa.array([False, None]),
]

batch0 = pa.record_batch(data0, names=names)
batch1 = pa.record_batch(data1, names=names)
batch2 = pa.record_batch(data2, names=names)


@pytest.fixture
def data_source_from_init_storage() -> Callable[[str, int], DataSource[TableStructure]]:
    def _data_source_from_init_storage(
        data_uri: str, num_partitions: int
    ) -> DataSource[TableStructure]:
        table = pa.Table.from_arrays(data0, names)
        structure = TableStructure.from_arrow_table(table, npartitions=num_partitions)
        data_source = DataSource(
            management=Management.writable,
            mimetype="application/x-tiled-sql-table",
            structure_family=StructureFamily.table,
            structure=structure,
            assets=[],
        )

        storage = cast(SQLStorage, parse_storage(data_uri))
        register_storage(storage)
        return SQLAdapter.init_storage(data_source=data_source, storage=storage)

    return _data_source_from_init_storage


@pytest.fixture
def adapter_duckdb_one_partition(
    tmp_path: Path,
    data_source_from_init_storage: Callable[[str, int], DataSource[TableStructure]],
    duckdb_uri: str,
) -> Generator[SQLAdapter, None, None]:
    data_source = data_source_from_init_storage(duckdb_uri, 1)
    yield SQLAdapter(
        data_source.assets[0].data_uri,
        data_source.structure,
        data_source.parameters["table_name"],
        data_source.parameters["dataset_id"],
    )


@pytest.fixture
def adapter_duckdb_many_partitions(
    tmp_path: Path,
    data_source_from_init_storage: Callable[[str, int], DataSource[TableStructure]],
    duckdb_uri: str,
) -> Generator[SQLAdapter, None, None]:
    data_source = data_source_from_init_storage(duckdb_uri, 3)
    yield SQLAdapter(
        data_source.assets[0].data_uri,
        data_source.structure,
        data_source.parameters["table_name"],
        data_source.parameters["dataset_id"],
    )


def test_attributes_duckdb_one_part(adapter_duckdb_one_partition: SQLAdapter) -> None:
    assert adapter_duckdb_one_partition.structure().columns == names
    assert adapter_duckdb_one_partition.structure().npartitions == 1


def test_attributes_duckdb_many_part(
    adapter_duckdb_many_partitions: SQLAdapter,
) -> None:
    assert adapter_duckdb_many_partitions.structure().columns == names
    assert adapter_duckdb_many_partitions.structure().npartitions == 3


@pytest.fixture
def adapter_sqlite_one_partition(
    tmp_path: Path,
    data_source_from_init_storage: Callable[[str, int], DataSource[TableStructure]],
    sqlite_uri: str,
) -> Generator[SQLAdapter, None, None]:
    data_source = data_source_from_init_storage(sqlite_uri, 1)
    yield SQLAdapter(
        data_source.assets[0].data_uri,
        data_source.structure,
        data_source.parameters["table_name"],
        data_source.parameters["dataset_id"],
    )


@pytest.fixture
def adapter_sqlite_many_partitions(
    tmp_path: Path,
    data_source_from_init_storage: Callable[[str, int], DataSource[TableStructure]],
    sqlite_uri: str,
) -> Generator[SQLAdapter, None, None]:
    data_source = data_source_from_init_storage(sqlite_uri, 3)
    yield SQLAdapter(
        data_source.assets[0].data_uri,
        data_source.structure,
        data_source.parameters["table_name"],
        data_source.parameters["dataset_id"],
    )


def test_attributes_sql_one_part(adapter_sqlite_one_partition: SQLAdapter) -> None:
    assert adapter_sqlite_one_partition.structure().columns == names
    assert adapter_sqlite_one_partition.structure().npartitions == 1


def test_attributes_sql_many_part(adapter_sqlite_many_partitions: SQLAdapter) -> None:
    assert adapter_sqlite_many_partitions.structure().columns == names
    assert adapter_sqlite_many_partitions.structure().npartitions == 3


@pytest.fixture
def adapter_psql_one_partition(
    data_source_from_init_storage: Callable[[str, int], DataSource[TableStructure]],
    postgres_uri: str,
) -> Generator[SQLAdapter, None, None]:
    data_source = data_source_from_init_storage(postgres_uri, 1)
    yield SQLAdapter(
        data_source.assets[0].data_uri,
        data_source.structure,
        data_source.parameters["table_name"],
        data_source.parameters["dataset_id"],
    )

    # Close all connections and dispose of the storage
    storage = get_storage(sanitize_uri(postgres_uri)[0])
    cast(SQLStorage, storage).dispose()


@pytest.fixture
def adapter_psql_many_partitions(
    data_source_from_init_storage: Callable[[str, int], DataSource[TableStructure]],
    postgres_uri: str,
) -> Generator[SQLAdapter, None, None]:
    data_source = data_source_from_init_storage(postgres_uri, 3)
    yield SQLAdapter(
        data_source.assets[0].data_uri,
        data_source.structure,
        data_source.parameters["table_name"],
        data_source.parameters["dataset_id"],
    )

    # Close all connections and dispose of the storage
    storage = get_storage(sanitize_uri(postgres_uri)[0])
    cast(SQLStorage, storage).dispose()


def test_psql(adapter_psql_one_partition: SQLAdapter) -> None:
    assert adapter_psql_one_partition.structure().columns == names
    assert adapter_psql_one_partition.structure().npartitions == 1


@pytest.mark.parametrize(
    "adapter",
    [
        ("adapter_sqlite_one_partition"),
        ("adapter_duckdb_one_partition"),
        ("adapter_psql_one_partition"),
    ],
)
def test_write_read_one_batch_one_part(
    adapter: SQLAdapter, request: pytest.FixtureRequest
) -> None:
    # get adapter from fixture
    adapter = request.getfixturevalue(adapter)

    # test appending and reading a table as a whole
    test_table = pa.Table.from_arrays(data0, names)

    adapter.append_partition(batch0, 0)
    result_read = adapter.read()
    # the pandas dataframe gives the last column of the data as 0 and 1 since SQL does not save boolean
    # so we explicitely convert the last column to boolean for testing purposes
    result_read["f3"] = result_read["f3"].astype("boolean")
    assert test_table == pa.Table.from_pandas(result_read)

    # test appending and reading a partition in a table
    result_read_partition = adapter.read_partition(0)
    # the pandas dataframe gives the last column of the data as 0 and 1 since SQL does not save boolean
    # so we explicitely convert the last column to boolean for testing purposes
    result_read_partition["f3"] = result_read_partition["f3"].astype("boolean")
    assert test_table == pa.Table.from_pandas(result_read_partition)


@pytest.mark.parametrize(
    "adapter",
    [
        ("adapter_sqlite_one_partition"),
        ("adapter_duckdb_one_partition"),
        ("adapter_psql_one_partition"),
    ],
)
def test_write_read_list_batch_one_part(
    adapter: SQLAdapter, request: pytest.FixtureRequest
) -> None:
    # get adapter from fixture
    adapter = request.getfixturevalue(adapter)

    test_table = pa.Table.from_batches([batch0, batch1, batch2])
    # test appending a list of batches to a table and read as a whole
    adapter.append_partition([batch0, batch1, batch2], 0)
    result_read = adapter.read()

    result_read["f3"] = result_read["f3"].astype("boolean")
    assert test_table == pa.Table.from_pandas(result_read)

    # test appending and reading a partition in a table
    result_read_partition = adapter.read_partition(0)

    result_read_partition["f3"] = result_read_partition["f3"].astype("boolean")
    assert test_table == pa.Table.from_pandas(result_read_partition)

    # test appending few more times done correctly
    test_table = pa.Table.from_batches(
        [batch0, batch1, batch2, batch2, batch0, batch1, batch1, batch2, batch0]
    )
    adapter.append_partition([batch2, batch0, batch1], 0)
    adapter.append_partition([batch1, batch2, batch0], 0)
    result_read = adapter.read()

    result_read["f3"] = result_read["f3"].astype("boolean")

    assert test_table == pa.Table.from_pandas(result_read)

    # test appending a few times and reading done correctly
    result_read_partition = adapter.read_partition(0)

    result_read_partition["f3"] = result_read_partition["f3"].astype("boolean")

    assert test_table == pa.Table.from_pandas(result_read_partition)


def assert_same_rows(table1: pa.Table, table2: pa.Table) -> None:
    "Verify that two tables have the same rows, regardless of order."
    assert table1.num_rows == table2.num_rows

    rows1 = {tuple(row) for row in table1.to_pylist()}
    rows2 = {tuple(row) for row in table2.to_pylist()}
    assert rows1 == rows2


@pytest.mark.parametrize(
    "adapter",
    [
        ("adapter_sqlite_many_partitions"),
        ("adapter_duckdb_many_partitions"),
        ("adapter_psql_many_partitions"),
    ],
)
def test_append_single_partition(
    adapter: SQLAdapter, request: pytest.FixtureRequest
) -> None:
    # get adapter from fixture
    adapter = request.getfixturevalue(adapter)

    # test writing an entire pyarrow table to a single partition
    table = pa.Table.from_batches([batch0, batch1, batch2])
    adapter.append_partition(table, 0)

    result_read = adapter.read()
    result_read["f3"] = result_read["f3"].astype("boolean")
    assert table == pa.Table.from_pandas(result_read)

    # test reading a specific partition
    result_read_partition = adapter.read_partition(0)
    result_read_partition["f3"] = result_read_partition["f3"].astype("boolean")
    assert table == pa.Table.from_pandas(result_read_partition)


@pytest.mark.parametrize(
    "adapter",
    [
        ("adapter_sqlite_many_partitions"),
        ("adapter_psql_many_partitions"),
    ],
)
def test_write_read_one_batch_many_part(
    adapter: SQLAdapter, request: pytest.FixtureRequest
) -> None:
    # get adapter from fixture
    adapter = request.getfixturevalue(adapter)

    # test writing to many partitions and reading it whole
    adapter.append_partition(batch0, 0)
    adapter.append_partition(batch1, 1)
    adapter.append_partition(batch2, 2)

    result_read = adapter.read()
    result_read["f3"] = result_read["f3"].astype("boolean")

    assert pa.Table.from_batches([batch0, batch1, batch2]) == pa.Table.from_pandas(
        result_read
    )

    # test reading a specific partition
    result_read_partition = adapter.read_partition(0)
    result_read_partition["f3"] = result_read_partition["f3"].astype("boolean")
    assert pa.Table.from_arrays(data0, names) == pa.Table.from_pandas(
        result_read_partition
    )

    result_read_partition = adapter.read_partition(1)
    result_read_partition["f3"] = result_read_partition["f3"].astype("boolean")
    assert pa.Table.from_arrays(data1, names) == pa.Table.from_pandas(
        result_read_partition
    )

    result_read_partition = adapter.read_partition(2)
    result_read_partition["f3"] = result_read_partition["f3"].astype("boolean")
    assert pa.Table.from_arrays(data2, names) == pa.Table.from_pandas(
        result_read_partition
    )

    # test appending a few times and reading done correctly
    adapter.append_partition(batch0, 1)
    adapter.append_partition(batch1, 2)
    adapter.append_partition(batch2, 0)

    result_read = adapter.read()
    result_read["f3"] = result_read["f3"].astype("boolean")

    # Check that each partition matches
    assert_same_rows(
        pa.Table.from_batches([batch0, batch2]),
        pa.Table.from_pandas(adapter.read_partition(0)),
    )
    assert_same_rows(
        pa.Table.from_batches([batch1, batch0]),
        pa.Table.from_pandas(adapter.read_partition(1)),
    )
    assert_same_rows(
        pa.Table.from_batches([batch2, batch1]),
        pa.Table.from_pandas(adapter.read_partition(2)),
    )
    assert_same_rows(
        pa.Table.from_batches([batch0, batch2, batch1, batch0, batch2, batch1]),
        pa.Table.from_pandas(result_read),
    )

    # read a specific field
    result_read = adapter.read_partition(0, fields=["f1"])
    assert [*data0[1].tolist(), *data2[1].tolist()] == result_read["f1"].tolist()
    result_read = adapter.read_partition(1, fields=["f0"])
    assert [*data1[0].tolist(), *data0[0].tolist()] == result_read["f0"].tolist()
    result_read = adapter.read_partition(2, fields=["f2"])
    assert [*data2[2].tolist(), *data1[2].tolist()] == result_read["f2"].tolist()


@pytest.mark.parametrize(
    "table_name, expected",
    [
        (
            "table_abcdefg12423pnjsbldfhjdfbv_hbdhfljb128w40_ndgjfsdflfnscljm",
            pytest.raises(
                ValueError, match=r"Invalid SQL identifier.+max character number is 63"
            ),
        ),
        (
            "create_abcdefg12423pnjsbldfhjdfbv_hbdhfljb128w40_ndgjfsdflfnscljk_sdbf_jhvjkbefl",
            pytest.raises(
                ValueError, match=r"Invalid SQL identifier.+max character number is 63"
            ),
        ),
        (
            "hello_abcdefg12423pnjsbldfhjdfbv_hbdhfljb128w40_ndgjfsdflfnscljk_sdbf_jhvjkbefl",
            pytest.raises(
                ValueError, match=r"Invalid SQL identifier.+max character number is 63"
            ),
        ),
        ("my_table_here_123_", None),
        ("the_short_table12374620_hello_table23704ynnm", None),
        (
            "_here_is_my_table",
            pytest.raises(ValueError, match=r"Malformed SQL identifier.+"),
        ),
        (
            "here-is-my-table",
            pytest.raises(ValueError, match=r"Malformed SQL identifier.+"),
        ),
        (
            "-here-is-my-table",
            pytest.raises(ValueError, match=r"Malformed SQL identifier.+"),
        ),
        (
            "here-is-my-table-",
            pytest.raises(ValueError, match=r"Malformed SQL identifier.+"),
        ),
        (
            "create_this_table1246*",
            pytest.raises(ValueError, match=r"Malformed SQL identifier.+"),
        ),
        (
            "create this_table1246",
            pytest.raises(ValueError, match=r"Malformed SQL identifier.+"),
        ),
        (
            "drop this_table1246",
            pytest.raises(ValueError, match=r"Malformed SQL identifier.+"),
        ),
        (
            "table_mytable!",
            pytest.raises(ValueError, match=r"Malformed SQL identifier.+"),
        ),
        ("my_table_here_123_", None),
        ("the_short_table12374620_hello_table23704ynnm", None),
        (
            "select",
            pytest.raises(
                ValueError,
                match=r"Reserved SQL keywords are not allowed in identifiers.+",
            ),
        ),
        (
            "create",
            pytest.raises(
                ValueError,
                match=r"Reserved SQL keywords are not allowed in identifiers.+",
            ),
        ),
        (
            "SELECT",
            pytest.raises(
                ValueError,
                match=r"Reserved SQL keywords are not allowed in identifiers.+",
            ),
        ),
        (
            "from",
            pytest.raises(
                ValueError,
                match=r"Reserved SQL keywords are not allowed in identifiers.+",
            ),
        ),
        ("drop_this_table123_", None),
        (
            "DROP_thistable123_hwejk",
            pytest.raises(ValueError, match=r"Malformed SQL identifier.+"),
        ),
        (
            "CAPITALIZED_NAME",
            pytest.raises(ValueError, match=r"Malformed SQL identifier.+"),
        ),
    ],
)
def test_check_table_name_is_safe(table_name: str, expected: Union[None, Any]) -> None:
    if isinstance(expected, type(pytest.raises(ValueError))):
        with expected:
            is_safe_identifier(
                table_name, TABLE_NAME_PATTERN, allow_reserved_words=False
            )
    else:
        assert is_safe_identifier(
            table_name, TABLE_NAME_PATTERN, allow_reserved_words=False
        )


@pytest.mark.parametrize(
    "column_name, expected",
    [
        # Valid column names
        ("valid_column_name", None),
        ("_another_valid_name123", None),
        ("column_name_with_underscores", None),
        ("COLUMNnameWITHCAPITALletters", None),
        ("short", None),
        ("a" * 63, None),  # Maximum length
        ("name with-other*allowed:special/characters?!", None),
        # Invalid identifiers - length
        (
            "a" * 64,
            pytest.raises(
                ValueError, match=r"Invalid SQL identifier.+max character number is 63"
            ),
        ),
        # Invalid identifiers - malformed
        (
            "1invalid_start",
            pytest.raises(ValueError, match=r"Malformed SQL identifier.+"),
        ),
        ("-invalid", pytest.raises(ValueError, match=r"Malformed SQL identifier.+")),
        # Invalid identifiers - forbidden characters
        (
            'invalid"name',
            pytest.raises(
                ValueError,
                match=r"Invalid SQL identifier.+contains forbidden character.+",
            ),
        ),
        (
            "invalid'name",
            pytest.raises(
                ValueError,
                match=r"Invalid SQL identifier.+contains forbidden character.+",
            ),
        ),
        (
            "invalid`name",
            pytest.raises(
                ValueError,
                match=r"Invalid SQL identifier.+contains forbidden character.+",
            ),
        ),
        (
            "invalid;name",
            pytest.raises(
                ValueError,
                match=r"Invalid SQL identifier.+contains forbidden character.+",
            ),
        ),
        (
            "invalid--name",
            pytest.raises(
                ValueError,
                match=r"Invalid SQL identifier.+contains forbidden character.+",
            ),
        ),
        (
            "invalid\\*name",
            pytest.raises(
                ValueError,
                match=r"Invalid SQL identifier.+contains forbidden character.+",
            ),
        ),
        (
            "invalid*\\name",
            pytest.raises(
                ValueError,
                match=r"Invalid SQL identifier.+contains forbidden character.+",
            ),
        ),
        (
            "invalid*/name",
            pytest.raises(
                ValueError,
                match=r"Invalid SQL identifier.+contains forbidden character.+",
            ),
        ),
        (
            "invalid\\name",
            pytest.raises(
                ValueError,
                match=r"Invalid SQL identifier.+contains forbidden character.+",
            ),
        ),
        (
            "invalid(name",
            pytest.raises(
                ValueError,
                match=r"Invalid SQL identifier.+contains forbidden character.+",
            ),
        ),
        (
            "invalid)name",
            pytest.raises(
                ValueError,
                match=r"Invalid SQL identifier.+contains forbidden character.+",
            ),
        ),
        (
            "invalid+name",
            pytest.raises(
                ValueError,
                match=r"Invalid SQL identifier.+contains forbidden character.+",
            ),
        ),
        (
            "invalid=name",
            pytest.raises(
                ValueError,
                match=r"Invalid SQL identifier.+contains forbidden character.+",
            ),
        ),
        (
            "yet\\another--invalid=name+with(many)forbidden*/characters",
            pytest.raises(
                ValueError,
                match=r"Invalid SQL identifier.+contains forbidden character.+",
            ),
        ),
    ],
)
def test_check_column_name_is_safe(column_name: str, expected: str) -> None:
    if isinstance(expected, type(pytest.raises(ValueError))):
        with expected:
            is_safe_identifier(
                column_name, COLUMN_NAME_PATTERN, allow_reserved_words=True
            )
    else:
        assert is_safe_identifier(
            column_name, COLUMN_NAME_PATTERN, allow_reserved_words=True
        )


@pytest.mark.parametrize("data_uri", ["sqlite_uri", "duckdb_uri", "postgres_uri"])
@pytest.mark.parametrize("column_name", ["a", "a b", "a-b", "a:b", "a*b", "a/b", "Ab"])
def test_can_query_with_valid_column_names(
    data_uri: str, column_name: str, request: pytest.FixtureRequest
) -> None:
    table = pa.Table.from_arrays([[1, 2, 3]], [column_name])
    structure = TableStructure.from_arrow_table(table)
    data_source = DataSource(
        management=Management.writable,
        mimetype="application/x-tiled-sql-table",
        structure_family=StructureFamily.table,
        structure=structure,
        assets=[],
    )
    data_uri = request.getfixturevalue(data_uri)
    storage = cast(SQLStorage, parse_storage(data_uri))
    register_storage(storage)
    assert SQLAdapter.init_storage(data_source=data_source, storage=storage) is not None
    storage.dispose()


@pytest.mark.parametrize("data_uri", ["sqlite_uri", "duckdb_uri", "postgres_uri"])
def test_reject_colliding_uppercase_column_names(
    data_uri: str, request: pytest.FixtureRequest
) -> None:
    # Define a table and a storage
    data_uri = request.getfixturevalue(data_uri)
    storage = cast(SQLStorage, parse_storage(data_uri))
    register_storage(storage)

    # Create a table with colliding column names
    table = pa.Table.from_arrays([[1, 2, 3], [4, 5, 6]], ["column_name", "COLUMN_NAME"])
    structure = TableStructure.from_arrow_table(table)
    data_source = DataSource(
        management=Management.writable,
        mimetype="application/x-tiled-sql-table",
        structure_family=StructureFamily.table,
        structure=structure,
        parameters={"table_name": "table_name"},
        assets=[],
    )
    with pytest.raises(ValueError, match=r"Column names must be unique.+"):
        SQLAdapter.init_storage(data_source=data_source, storage=storage)

    # Create a table with mixed cases in column names
    table = pa.Table.from_arrays([[1, 2, 3], [4, 5, 6]], ["lower_case", "UPPER_CASE"])
    structure = TableStructure.from_arrow_table(table)
    data_source = DataSource(
        management=Management.writable,
        mimetype="application/x-tiled-sql-table",
        structure_family=StructureFamily.table,
        structure=structure,
        parameters={"table_name": "table_name"},
        assets=[],
    )
    data_source = SQLAdapter.init_storage(data_source=data_source, storage=storage)
    assert data_source is not None

    # Write to and read from the table
    adapter = SQLAdapter(
        data_source.assets[0].data_uri,
        structure=data_source.structure,
        table_name=data_source.parameters["table_name"],
        dataset_id=data_source.parameters["dataset_id"],
    )
    adapter.append_partition(table, 0)
    assert adapter.table_name == "table_name"
    assert set(adapter.read().columns) == {"lower_case", "UPPER_CASE"}

    storage.dispose()  # Close all connections


@pytest.mark.parametrize(
    "initial, appended",
    [
        ([1, 2, 3], [None, None, None]),
        ([1.5, 2.5, 3.5], [None, None, None]),
        (["a", "b", "c"], [None, None, None]),
        ([[1], [2, 4], [3]], [[], [], []]),
        ([[1], [2, 4], [3]], [None, None, None]),
        ([[1], [2, 4], [3]], [[None], [None], [None]]),
        ([[1.5], [2.5, 4.5], [3.5]], [[], [], []]),
        ([[1.5], [2.5, 4.5], [3.5]], [None, None, None]),
        ([[1.5], [2.5, 4.5], [3.5]], [[None], [None], [None]]),
        ([["a"], ["b1", "b2"], ["c"]], [[], [], []]),
        ([["a"], ["b1", "b2"], ["c"]], [[None], [None], [None]]),
        ([["a"], ["b1", "b2"], ["c"]], [None, None, None]),
    ],
)
@pytest.mark.parametrize("data_uri", ["sqlite_uri", "duckdb_uri", "postgres_uri"])
def test_append_nullable(
    initial: list[Any],
    appended: list[Any],
    data_uri: str,
    request: pytest.FixtureRequest,
) -> None:
    "Test appending nullable data in an SQL table and reading them back."

    if (data_uri == "sqlite_uri") and (isinstance(initial[0], list)):
        pytest.xfail(reason="Unsupported PyArrow type in SQLite")

    def deep_array_equal(a1: Any, a2: Any) -> bool:
        "Compare two (possibly nested) arrays for equality, including NaN values."
        if not (isinstance(a1, np.ndarray) and isinstance(a2, np.ndarray)):
            # Both are scalar values
            return bool((a1 == a2) or (np.isnan(a1) and np.isnan(a2)))
        elif (len(a1) == 0) and (len(a2) == 0):
            # Both are empty arrays
            return True
        elif len(a1) != len(a2):
            return False
        else:
            return all(deep_array_equal(x1, x2) for x1, x2 in zip(a1, a2))

    # Define a table and a storage
    data_uri = request.getfixturevalue(data_uri)
    storage = cast(SQLStorage, parse_storage(data_uri))
    register_storage(storage)

    # Create a table to be appended
    table_0 = pa.Table.from_arrays([initial], ["part_column"])
    data_source = DataSource(
        management=Management.writable,
        mimetype="application/x-tiled-sql-table",
        structure_family=StructureFamily.table,
        structure=TableStructure.from_arrow_table(table_0),
        parameters={"table_name": "part_table"},
        assets=[],
    )
    data_source = SQLAdapter.init_storage(data_source=data_source, storage=storage)

    # Write the first part of the data to the table
    adapter_part = SQLAdapter(
        data_source.assets[0].data_uri,
        structure=data_source.structure,
        table_name=data_source.parameters["table_name"],
        dataset_id=data_source.parameters["dataset_id"],
    )
    adapter_part.append_partition(table_0, 0)

    # Write the second part of the data to the table and read it back
    table_1 = pa.Table.from_arrays([appended], ["part_column"])
    adapter_part.append_partition(table_1, 0)
    result_part = adapter_part.read()["part_column"].to_numpy()

    # Write the full table at once and read it back
    table_full = pa.Table.from_arrays([initial + appended], ["full_column"])
    data_source = DataSource(
        management=Management.writable,
        mimetype="application/x-tiled-sql-table",
        structure_family=StructureFamily.table,
        structure=TableStructure.from_arrow_table(table_full),
        parameters={"table_name": "full_table"},
        assets=[],
    )
    data_source = SQLAdapter.init_storage(data_source=data_source, storage=storage)
    adapter_full = SQLAdapter(
        data_source.assets[0].data_uri,
        structure=data_source.structure,
        table_name=data_source.parameters["table_name"],
        dataset_id=data_source.parameters["dataset_id"],
    )
    adapter_full.append_partition(table_full, 0)
    result_full = adapter_full.read()["full_column"].to_numpy()

    # Check if the data matches in both cases
    assert deep_array_equal(result_part, result_full)

    storage.dispose()  # Close all connections



import anyio
import h5py
import numpy as np
import pandas
import pytest
import zarr

from ..adapters.zarr import ZARR_LIB_V2
from ..catalog import in_memory
from ..client import Context, from_context, record_history
from ..client.register import register
from ..examples.generate_files import generate_files
from ..server.app import build_app


@pytest.fixture(scope="module")
def client(request: pytest.FixtureRequest):
    "A tiled client that serves from parametrized data_dir (Path or str)"
    data_dir = request.getfixturevalue(request.param)
    catalog = in_memory(readable_storage=[data_dir])
    with Context.from_app(build_app(catalog)) as context:
        client = from_context(context)
        anyio.run(register, client, data_dir)
        yield client


@pytest.fixture(scope="module")
def example_data_dir(tmpdir_factory):
    "Generate a temporary directory with example files."
    tmpdir = tmpdir_factory.mktemp("example_files")
    generate_files(tmpdir)
    return tmpdir


@pytest.mark.parametrize("fields", (None, (), ("a", "b")))
@pytest.mark.parametrize("client", ("example_data_dir",), indirect=True)
def test_directory_fields(client, fields, buffer):
    "Export selected fields (files) from a directory via /container/full."
    url_path = client.item["links"]["full"]
    with record_history() as history:
        client.export(buffer, fields=fields, format="application/x-hdf5")

    assert_single_request_to_url(history, url_path)
    assert_requested_fields_fetched(buffer, fields, client)


@pytest.fixture(scope="module")
def excel_data_dir(tmpdir_factory):
    "Generate a temporary Excel file with multiple Sheets of tabular data."
    tmpdir = tmpdir_factory.mktemp("excel_files")
    df = pandas.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    with pandas.ExcelWriter(tmpdir / "spreadsheet.xlsx") as writer:
        for i in range(10):
            df.to_excel(writer, sheet_name=f"Sheet {i+1}", index=False)
    return tmpdir


@pytest.mark.parametrize("fields", (None, (), ("Sheet 1", "Sheet 10")))
@pytest.mark.parametrize("client", ("excel_data_dir",), indirect=True)
def test_excel_fields(client, fields, buffer):
    "Export selected fields (sheets) from an Excel file via /container/full."
    client = client["spreadsheet"]
    url_path = client.item["links"]["full"]
    with record_history() as history:
        client.export(buffer, fields=fields, format="application/x-hdf5")
        # TODO: Enable container to export XLSX if all nodes are tables?
        # client.export(buffer, fields=fields, format=XLSX_MIME_TYPE)

    assert_single_request_to_url(history, url_path)
    assert_requested_fields_fetched(buffer, fields, client)


def mark_xfail(value, unsupported="UNSPECIFIED ADAPTER"):
    "Indicate that this parameterized value is expected to fail"
    reason = f"Tiled does not currently support selecting 'fields' for {unsupported}"
    return pytest.param(value, marks=pytest.mark.xfail(reason=reason))


@pytest.fixture(scope="module")
def zarr_data_dir(tmpdir_factory):
    "Generate a temporary Zarr group file with multiple datasets."
    tmpdir = tmpdir_factory.mktemp("zarr_files")
    try:
        root = zarr.open(str(tmpdir / "zarr_group.zarr"), mode="w")
        for i, name in enumerate("abcde"):
            if ZARR_LIB_V2:
                root.create_dataset(name, data=np.arange(i, i + 3))
            else:
                root.create_array(name, data=np.arange(i, i + 3))
    finally:
        # Ensure the Zarr group is closed properly
        print("Closed")
    return tmpdir


@pytest.mark.parametrize("fields", (None, (), mark_xfail(("b", "d"), "Zarr")))
@pytest.mark.parametrize("client", ("zarr_data_dir",), indirect=True)
def test_zarr_group_fields(client, fields, buffer):
    "Export selected fields (Datasets) from a Zarr group via /container/full."
    client = client["zarr_group"]
    # Normally, zarr would have 'attributes' stored as internal dictionary in the
    # metadata, but HDF5 does not support nested dictionaries.
    client.replace_metadata({"attributes": "", "zarr_format": 2 if ZARR_LIB_V2 else 3})
    url_path = client.item["links"]["full"]
    with record_history() as history:
        client.export(buffer, fields=fields, format="application/x-hdf5")

        assert_single_request_to_url(history, url_path)
        assert_requested_fields_fetched(buffer, fields, client)


@pytest.fixture(scope="module")
def hdf5_data_dir(tmpdir_factory):
    "Generate a temporary HDF5 file with multiple entries."
    tmpdir = tmpdir_factory.mktemp("hdf5_files")
    with h5py.File(str(tmpdir / "hdf5_example.h5"), "w") as file:
        file["x"] = [1, 2, 3]
        group = file.create_group("g")
        group["y"] = [4, 5, 6]
    return tmpdir


@pytest.mark.parametrize(
    "fields", (None, (), mark_xfail(("x",), "HDF5"), mark_xfail(("g",), "HDF5"))
)
@pytest.mark.parametrize("client", ("hdf5_data_dir",), indirect=True)
def test_hdf5_fields(client, fields, buffer):
    "Export selected fields (array/group) from a HDF5 file via /container/full."
    client = client["hdf5_example"]
    url_path = client.item["links"]["full"]
    with record_history() as history:
        client.export(buffer, fields=fields, format="application/x-hdf5")

    assert_single_request_to_url(history, url_path)
    assert_requested_fields_fetched(buffer, fields, client)


def assert_single_request_to_url(history, url_path):
    "Container contents were fetched as a single request using 'full' route."
    (request,) = history.requests
    request_url_path = request.url.copy_with(query=None)
    assert request_url_path == url_path


def assert_requested_fields_fetched(buffer, fields, client):
    "Only the requested fields were fetched."
    with h5py.File(buffer) as file:
        actual_fields = set(file.keys())
    expected = set(fields or client.keys())  # By default all fields were fetched
    assert actual_fields == expected



"""Add access_blob column to nodes

Revision ID: 9331ed94d6ac
Revises: 45a702586b2a
Create Date: 2025-05-07 12:44:45.886279

"""
import sqlalchemy as sa
from alembic import op

from tiled.catalog.orm import JSONVariant

# revision identifiers, used by Alembic.
revision = "9331ed94d6ac"
down_revision = "45a702586b2a"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "nodes",
        sa.Column("access_blob", JSONVariant, nullable=False, server_default="{}"),
    )


def downgrade():
    op.drop_column("nodes", "access_blob")



import asyncio
import collections
import contextvars
import logging
import os
import secrets
import sys
import urllib.parse
import warnings
from contextlib import asynccontextmanager
from functools import cache, partial
from pathlib import Path
from textwrap import dedent
from typing import Any, Callable, Coroutine, Optional, TypedDict, Union

import anyio
import packaging.version
import yaml
from asgi_correlation_id import CorrelationIdMiddleware, correlation_id
from fastapi import FastAPI, HTTPException, Request, Response
from fastapi.exception_handlers import http_exception_handler
from fastapi.middleware.cors import CORSMiddleware
from fastapi.openapi.utils import get_openapi
from fastapi.responses import HTMLResponse, JSONResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from starlette.middleware.base import RequestResponseEndpoint
from starlette.responses import FileResponse
from starlette.status import (
    HTTP_200_OK,
    HTTP_400_BAD_REQUEST,
    HTTP_401_UNAUTHORIZED,
    HTTP_403_FORBIDDEN,
    HTTP_404_NOT_FOUND,
    HTTP_409_CONFLICT,
    HTTP_500_INTERNAL_SERVER_ERROR,
)

from tiled.query_registration import QueryRegistry, default_query_registry
from tiled.server.protocols import ExternalAuthenticator, InternalAuthenticator

from ..catalog.adapter import WouldDeleteData
from ..config import construct_build_app_kwargs, parse_configs
from ..media_type_registration import (
    CompressionRegistry,
    SerializationRegistry,
    default_compression_registry,
    default_deserialization_registry,
    default_serialization_registry,
)
from ..utils import SHARE_TILED_PATH, Conflicts, UnsupportedQueryType
from ..validation_registration import ValidationRegistry, default_validation_registry
from .compression import CompressionMiddleware
from .router import get_router
from .settings import Settings, get_settings
from .utils import API_KEY_COOKIE_NAME, CSRF_COOKIE_NAME, get_root_url, record_timing
from .zarr import get_zarr_router_v2, get_zarr_router_v3

SAFE_METHODS = {"GET", "HEAD", "OPTIONS", "TRACE"}
SENSITIVE_COOKIES = {
    API_KEY_COOKIE_NAME,
}
CSRF_HEADER_NAME = "x-csrf"
CSRF_QUERY_PARAMETER = "csrf"

MINIMUM_SUPPORTED_PYTHON_CLIENT_VERSION = packaging.version.parse("0.1.0a104")

logger = logging.getLogger(__name__)
logger.setLevel("INFO")
handler = logging.StreamHandler()
handler.setLevel("DEBUG")
handler.setFormatter(logging.Formatter("%(message)s"))
logger.addHandler(handler)


# This is used to pass the currently-authenticated principal into the logger.
current_principal = contextvars.ContextVar("current_principal")


AppTask = Callable[[], Coroutine[None, None, Any]]
"""Async function to be run as part of the app's lifecycle"""


def custom_openapi(app):
    """
    The app's openapi method will be monkey-patched with this.

    This is the approach the documentation recommends.

    https://fastapi.tiangolo.com/advanced/extending-openapi/
    """
    from .. import __version__

    if app.openapi_schema:
        return app.openapi_schema
    # Customize heading.
    openapi_schema = get_openapi(
        title="Tiled",
        version=__version__,
        description="Structured data access service",
        routes=app.routes,
    )
    # Insert refreshUrl.
    openapi_schema["components"]["securitySchemes"]["OAuth2PasswordBearer"]["flows"][
        "password"
    ]["refreshUrl"] = "token/refresh"
    app.openapi_schema = openapi_schema
    return app.openapi_schema


def build_app(
    tree,
    authentication=None,
    server_settings=None,
    query_registry: Optional[QueryRegistry] = None,
    serialization_registry: Optional[SerializationRegistry] = None,
    deserialization_registry: Optional[SerializationRegistry] = None,
    compression_registry: Optional[CompressionRegistry] = None,
    validation_registry: Optional[ValidationRegistry] = None,
    tasks: Optional[dict[str, list[AppTask]]] = None,
    scalable=False,
    access_policy=None,
):
    """
    Serve a Tree

    Parameters
    ----------
    tree : Tree
    authentication: dict, optional
        Dict of authentication configuration.
    authenticators: list, optional
        List of authenticator classes (one per support identity provider)
    server_settings: dict, optional
        Dict of other server configuration.
    access_policy:
        AccessPolicy object encoding rules for which users can see which entries.
    """
    authentication = authentication or {}
    authenticators: dict[str, Union[ExternalAuthenticator, InternalAuthenticator]] = {
        spec["provider"]: spec["authenticator"]
        for spec in authentication.get("providers", [])
    }
    server_settings = server_settings or {}
    query_registry = query_registry or default_query_registry
    serialization_registry = serialization_registry or default_serialization_registry
    deserialization_registry = (
        deserialization_registry or default_deserialization_registry
    )
    compression_registry = compression_registry or default_compression_registry
    validation_registry = validation_registry or default_validation_registry

    tasks: TaskMap = {
        "startup": (tasks or {}).get("startup", []),
        "background": (tasks or {}).get("background", []),
        "shutdown": (tasks or {}).get("shutdown", []),
    }
    # The tasks are collected at config-parsing time off of the sub-trees.
    # Collect the tasks off the root tree here, so that it works when
    # a single tree is passed to build_app(...) directly, as happens in the tests.
    tasks["startup"].extend(getattr(tree, "startup_tasks", []))
    tasks["background"].extend(getattr(tree, "background_tasks", []))
    tasks["shutdown"].extend(getattr(tree, "shutdown_tasks", []))

    if scalable:
        if authentication.get("providers"):
            # Even if the deployment allows public, anonymous access, secret
            # keys are needed to generate JWTs for any users that do log in.
            if (
                "secret_keys" not in authentication
                and "TILED_SECRET_KEYS" not in os.environ
            ):
                raise UnscalableConfig(
                    dedent(
                        """
                        In a scaled (multi-process) deployment, when Tiled is configured with an
                        Authenticator, secret keys must be provided via configuration like

                        authentication:
                          secret_keys:
                            - SECRET
                          ...

                        or via the environment variable TILED_SECRET_KEYS.\
                        """,
                    )
                )
            # Multi-user authentication requires a database. We cannot fall
            # back to the default of an in-memory SQLite database in a
            # horizontally scaled deployment.
            if not server_settings.get("database", {}).get("uri"):
                raise UnscalableConfig(
                    dedent(
                        """
                        In a scaled (multi-process) deployment, when Tiled is configured with an
                        Authenticator, a persistent database must be provided via configuration like

                        database:
                          uri: sqlite:////path/to/database.sqlite

                        """
                    )
                )
        else:
            # No authentication provider is configured, so no secret keys are
            # needed, but a single-user API key must be set.
            if not (
                ("single_user_api_key" in authentication)
                or ("TILED_SINGLE_USER_API_KEY" in os.environ)
            ):
                raise UnscalableConfig(
                    dedent(
                        """
                        In a scaled (multi-process) deployment, when Tiled is configured for
                        single-user access (i.e. without an Authenticator) a single-user API key must
                        be provided via configuration like

                        authentication:
                          single_user_api_key: SECRET
                          ...

                        or via the environment variable TILED_SINGLE_USER_API_KEY.\
                        """,
                    )
                )

    # If we reach here, the no configuration problems were found.

    @asynccontextmanager
    async def lifespan(app: FastAPI):
        "Manage lifespan events for each event loop that the app runs in"
        await startup_event()
        yield
        await shutdown_event()

    app = FastAPI(lifespan=lifespan)

    # Healthcheck for deployment to containerized systems, needs to preempt other responses.
    # Standardized for Kubernetes, but also used by other systems.
    @app.get("/healthz", status_code=200)
    async def healthz():
        return {"status": "ready"}

    if SHARE_TILED_PATH:
        # If the distribution includes static assets, serve UI routes.

        @app.get("/ui/{path:path}")
        async def ui(path):
            response = await lookup_file(path)
            return response

        async def lookup_file(path, try_app=True):
            if not path:
                path = "index.html"
            full_path = Path(SHARE_TILED_PATH, "ui", path)
            try:
                stat_result = await anyio.to_thread.run_sync(os.stat, full_path)
            except PermissionError:
                raise HTTPException(status_code=HTTP_401_UNAUTHORIZED)
            except FileNotFoundError:
                # This may be a URL that has meaning to the client-side application,
                # such as /ui//metadata/a/b/c.
                # Serve index.html and let the client-side application sort it out.
                if try_app:
                    response = await lookup_file("index.html", try_app=False)
                    return response
                raise HTTPException(status_code=HTTP_404_NOT_FOUND)
            except OSError:
                raise
            return FileResponse(
                full_path,
                stat_result=stat_result,
                status_code=HTTP_200_OK,
            )

        app.mount(
            "/static",
            StaticFiles(directory=Path(SHARE_TILED_PATH, "static")),
            name="ui",
        )
        templates = Jinja2Templates(Path(SHARE_TILED_PATH, "templates"))

        @app.get("/", response_class=HTMLResponse)
        async def index(
            request: Request,
        ):
            return templates.TemplateResponse(
                request,
                "index.html",
                {
                    # This is used to construct the link to the React UI.
                    "root_url": get_root_url(request),
                    # If defined, this adds a Binder link to the page.
                    "binder_link": os.getenv("TILED_BINDER_LINK"),
                },
            )

        TILED_UI_SETTINGS = os.getenv("TILED_UI_SETTINGS")
        if TILED_UI_SETTINGS is None:
            TILED_UI_SETTINGS = Path(
                SHARE_TILED_PATH, "static", "default_ui_settings.yml"
            )
        if TILED_UI_SETTINGS != "":
            # If "", the settings are being served some other way, such as by
            # nginx, perhaps because the API is served from a sub-path of this netloc.

            # The settings are YAML-formatted because that is more readable and supports
            # comments. But they are served as JSON because that is easy to deal with
            # on the client side.
            ui_settings = yaml.safe_load(Path(TILED_UI_SETTINGS).read_text())
            if root_path := server_settings.get("root_path", ""):
                ui_settings["api_url"] = f"{root_path}{ui_settings['api_url']}"

            @app.get("/tiled-ui-settings")
            async def tiled_ui_settings():
                return ui_settings

    @app.exception_handler(Conflicts)
    async def conflicts_exception_handler(request: Request, exc: Conflicts):
        message = exc.args[0]
        return JSONResponse(status_code=HTTP_409_CONFLICT, content={"detail": message})

    @app.exception_handler(UnsupportedQueryType)
    async def unsupported_query_type_exception_handler(
        request: Request, exc: UnsupportedQueryType
    ):
        query_type = exc.args[0]
        return JSONResponse(
            status_code=HTTP_400_BAD_REQUEST,
            content={
                "detail": f"The query type {query_type!r} is not supported on this node."
            },
        )

    @app.exception_handler(WouldDeleteData)
    async def would_delete_data_exception_handler(
        request: Request, exc: WouldDeleteData
    ):
        return JSONResponse(
            status_code=HTTP_409_CONFLICT,
            content={"detail": exc.args[0]},
        )

    # This list will be mutated when settings are processed at app startup.
    app.state.allow_origins = []
    app.add_middleware(
        CORSMiddleware,
        allow_origins=app.state.allow_origins,
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
        # If we restrict the allowed_headers in future, remember to include
        # exemptions for these, related to asgi_correlation_id.
        # allow_headers=["X-Requested-With", "X-Tiled-Request-ID"],
        expose_headers=["X-Tiled-Request-ID"],
    )

    @app.exception_handler(Exception)
    async def unhandled_exception_handler(
        request: Request, exc: Exception
    ) -> JSONResponse:
        return await http_exception_handler(
            request,
            HTTPException(
                HTTP_500_INTERNAL_SERVER_ERROR,
                "Internal server error",
                headers={"X-Tiled-Request-ID": correlation_id.get() or ""},
            ),
        )

    router = get_router(
        query_registry,
        serialization_registry,
        deserialization_registry,
        validation_registry,
        authenticators,
    )
    app.include_router(router, prefix="/api/v1")

    app.include_router(get_zarr_router_v2(), prefix="/zarr/v2")
    app.include_router(get_zarr_router_v3(), prefix="/zarr/v3")

    # The Tree and Authenticator have the opportunity to add custom routes to
    # the server here. (Just for example, a Tree of BlueskyRuns uses this
    # hook to add a /documents route.) This has to be done before dependency_overrides
    # are processed, so we cannot just inject this configuration via Depends.
    for custom_router in getattr(tree, "include_routers", []):
        app.include_router(custom_router, prefix="/api/v1")

    app.state.access_policy = access_policy

    if authentication.get("providers", []):
        # Delay this imports to avoid delaying startup with the SQL and cryptography
        # imports if they are not needed.
        from .authentication import (
            add_external_routes,
            add_internal_routes,
            authentication_router,
            oauth2_scheme,
        )

        # For the OpenAPI schema, inject a OAuth2PasswordBearer URL.
        first_provider = authentication["providers"][0]["provider"]
        oauth2_scheme.model.flows.password.tokenUrl = (
            f"/api/v1/auth/provider/{first_provider}/token"
        )
        # Authenticators provide Router(s) for their particular flow.
        # Collect them in the authentication_router.
        authentication_router = authentication_router()
        # This adds the universal routes like /session/refresh and /session/revoke.
        # Below we will add routes specific to our authentication providers.

        for spec in authentication["providers"]:
            provider = spec["provider"]
            authenticator = spec["authenticator"]
            if isinstance(authenticator, InternalAuthenticator):
                add_internal_routes(authentication_router, provider, authenticator)
            elif isinstance(authenticator, ExternalAuthenticator):
                add_external_routes(authentication_router, provider, authenticator)
            else:
                raise ValueError(f"unknown authenticator type {type(authenticator)}")
            for custom_router in getattr(authenticator, "include_routers", []):
                authentication_router.include_router(
                    custom_router, prefix=f"/provider/{provider}"
                )
        # And add this authentication_router itself to the app.
        app.include_router(authentication_router, prefix="/api/v1/auth")
        app.state.authenticated = True
    else:
        app.state.authenticated = False

    # Expose the root_tree here to make it accessible from endpoints via
    # request.app.state and in tests via client.context.app.state
    app.state.root_tree = tree

    @cache
    def override_get_settings():
        settings = get_settings()
        for item in [
            "allow_anonymous_access",
            "secret_keys",
            "single_user_api_key",
            "access_token_max_age",
            "refresh_token_max_age",
            "session_max_age",
        ]:
            if authentication.get(item) is not None:
                setattr(settings, item, authentication[item])
        for item in [
            "allow_origins",
            "response_bytesize_limit",
            "exact_count_limit",
            "reject_undeclared_specs",
            "expose_raw_assets",
        ]:
            if server_settings.get(item) is not None:
                setattr(settings, item, server_settings[item])
        database = server_settings.get("database", {})
        if uri := database.get("uri"):
            settings.database_settings.uri = uri
        if pool_size := database.get("pool_size"):
            settings.database_settings.pool_size = pool_size
        if pool_pre_ping := database.get("pool_pre_ping"):
            settings.database_settings.pool_pre_ping = pool_pre_ping
        if max_overflow := database.get("max_overflow"):
            settings.database_settings.max_overflow = max_overflow
        if init_if_not_exists := database.get("init_if_not_exists"):
            settings.database_init_if_not_exists = init_if_not_exists
        if authentication.get("providers"):
            # If we support authentication providers, we need a database, so if one is
            # not set, use a SQLite database in memory. Horizontally scaled deployments
            # must specify a persistent database.
            settings.database_settings.uri = (
                settings.database_settings.uri or "sqlite://"
            )
        return settings

    async def startup_event():
        from .. import __version__

        logger.info(f"Tiled version {__version__}")
        # Validate the single-user API key.
        settings: Settings = app.dependency_overrides[get_settings]()
        single_user_api_key = settings.single_user_api_key
        API_KEY_MSG = dedent(
            """
            Here are two ways to generate a good API key:

            # With openssl:
            openssl rand -hex 32

            # With Python:
            python -c "import secrets; print(secrets.token_hex(32))"

            """
        )
        if single_user_api_key is not None:
            if not single_user_api_key:
                raise ValueError(
                    dedent(
                        """
                        The single-user API key is set to an empty value. Perhaps the environment
                        variable TILED_SINGLE_USER_API_KEY is set to an empty string.
                        """
                        + API_KEY_MSG
                    )
                )
            if not single_user_api_key.isalnum():
                raise ValueError(
                    dedent(
                        """
                        The API key must only contain alphanumeric characters. We enforce this because
                        pasting other characters into a URL, as in ?api_key=..., can result in
                        confusing behavior due to ambiguous encodings.
                        """
                    )
                    + API_KEY_MSG
                )

        # Run startup tasks collected from trees (adapters).
        for task in tasks["startup"]:
            await task()

        # Stash these to cancel this on shutdown.
        app.state.tasks = []
        # Trees and Authenticators can run tasks in the background.
        background_tasks: list[AppTask] = []
        background_tasks.extend(tasks["background"])
        for authenticator in authenticators:
            background_tasks.extend(getattr(authenticator, "background_tasks", []))
        for task in background_tasks:
            asyncio_task = asyncio.create_task(task())
            app.state.tasks.append(asyncio_task)

        app.state.allow_origins.extend(settings.allow_origins)

        if settings.database_settings.uri is not None:
            from sqlalchemy.ext.asyncio import AsyncSession

            from ..alembic_utils import (
                DatabaseUpgradeNeeded,
                UninitializedDatabase,
                check_database,
            )
            from ..authn_database import orm
            from ..authn_database.connection_pool import open_database_connection_pool
            from ..authn_database.core import (
                ALL_REVISIONS,
                REQUIRED_REVISION,
                initialize_database,
                make_admin_by_identity,
                purge_expired,
            )

            # This creates a connection pool and stashes it in a module-global
            # registry, keyed on database_settings, where can be retrieved by
            # the Dependency get_database_session.
            engine = open_database_connection_pool(settings.database_settings)
            if not engine.url.database or engine.url.query.get("mode") == "memory":
                # Special-case for in-memory SQLite: Because it is transient we can
                # skip over anything related to migrations.
                await initialize_database(engine)
                logger.info("Transient in-memory database initialized.")
            else:
                redacted_url = engine.url._replace(password="[redacted]")
                try:
                    await check_database(engine, REQUIRED_REVISION, ALL_REVISIONS)
                except UninitializedDatabase:
                    if settings.database_init_if_not_exists:
                        # The alembic stamping can only be does synchronously.
                        # The cleanest option available is to start a subprocess
                        # because SQLite is allergic to threads.
                        import subprocess

                        # TODO Check if catalog exists.
                        subprocess.run(
                            [
                                sys.executable,
                                "-m",
                                "tiled",
                                "admin",
                                "initialize-database",
                                str(engine.url),
                            ],
                            capture_output=True,
                            check=True,
                        )
                    else:
                        print(
                            dedent(
                                f"""

                                No database found at {redacted_url}

                                To create one, run:

                                    tiled admin init-database {redacted_url}
                                """
                            ),
                            file=sys.stderr,
                        )
                        raise
                except DatabaseUpgradeNeeded as err:
                    print(
                        dedent(
                            f"""

                            The database used by Tiled to store authentication-related information
                            was created using an older version of Tiled. It needs to be upgraded to
                            work with this version of Tiled.

                            Back up the database, and then run:

                                tiled admin upgrade-database {redacted_url}
                            """
                        ),
                        file=sys.stderr,
                    )
                    raise err from None
                else:
                    logger.info(f"Connected to existing database at {redacted_url}.")
            for admin in authentication.get("tiled_admins", []):
                logger.info(
                    f"Ensuring that principal with identity {admin} has role 'admin'"
                )
                async with AsyncSession(
                    engine, autoflush=False, expire_on_commit=False
                ) as session:
                    await make_admin_by_identity(
                        session,
                        identity_provider=admin["provider"],
                        id=admin["id"],
                    )

            if app.state.access_policy is not None and hasattr(
                app.state.access_policy, "access_tags_parser"
            ):
                await app.state.access_policy.access_tags_parser.connect()

            async def purge_expired_sessions_and_api_keys():
                PURGE_INTERVAL = 600  # seconds
                while True:
                    async with AsyncSession(
                        engine, autoflush=False, expire_on_commit=False
                    ) as db_session:
                        num_expired_sessions = await purge_expired(
                            db_session, orm.Session
                        )
                        if num_expired_sessions:
                            logger.info(
                                f"Purged {num_expired_sessions} expired Sessions from the database."
                            )
                        num_expired_api_keys = await purge_expired(
                            db_session, orm.APIKey
                        )
                        if num_expired_api_keys:
                            logger.info(
                                f"Purged {num_expired_api_keys} expired API keys from the database."
                            )
                    await asyncio.sleep(PURGE_INTERVAL)

            app.state.tasks.append(
                asyncio.create_task(purge_expired_sessions_and_api_keys())
            )

    async def shutdown_event():
        # Run shutdown tasks collected from trees (adapters).
        for task in tasks["shutdown"]:
            await task()

        settings: Settings = app.dependency_overrides[get_settings]()
        if settings.database_settings.uri is not None:
            from ..authn_database.connection_pool import close_database_connection_pool

            await close_database_connection_pool(settings.database_settings)
        for task in app.state.tasks:
            task.cancel()

    app.add_middleware(
        CompressionMiddleware,
        compression_registry=compression_registry,
        minimum_size=1000,
    )

    @app.middleware("http")
    async def double_submit_cookie_csrf_protection(
        request: Request, call_next: RequestResponseEndpoint
    ):
        # https://cheatsheetseries.owasp.org/cheatsheets/Cross-Site_Request_Forgery_Prevention_Cheat_Sheet.html#double-submit-cookie
        csrf_cookie = request.cookies.get(CSRF_COOKIE_NAME)
        if (request.method not in SAFE_METHODS) and set(request.cookies).intersection(
            SENSITIVE_COOKIES
        ):
            if not csrf_cookie:
                return Response(
                    status_code=HTTP_403_FORBIDDEN,
                    content="Expected tiled_csrf_token cookie",
                )
            # Get the token from the Header or (if not there) the query parameter.
            csrf_token = request.headers.get(CSRF_HEADER_NAME)
            if csrf_token is None:
                parsed_query = urllib.parse.parse_qs(request.url.query)
                csrf_token = parsed_query.get(CSRF_QUERY_PARAMETER)
            if not csrf_token:
                return Response(
                    status_code=HTTP_403_FORBIDDEN,
                    content=f"Expected {CSRF_QUERY_PARAMETER} query parameter or {CSRF_HEADER_NAME} header",
                )
            # Securely compare the token with the cookie.
            if not secrets.compare_digest(csrf_token, csrf_cookie):
                return Response(
                    status_code=HTTP_403_FORBIDDEN,
                    content="Double-submit CSRF tokens do not match",
                )

        response = await call_next(request)
        if not csrf_cookie:
            response.set_cookie(
                key=CSRF_COOKIE_NAME,
                value=secrets.token_urlsafe(32),
                httponly=True,
                samesite="lax",
            )
        return response

    @app.middleware("http")
    async def client_compatibility_check(
        request: Request, call_next: RequestResponseEndpoint
    ):
        user_agent = request.headers.get("user-agent", "")
        if user_agent.startswith("python-tiled/"):
            agent, _, raw_version = user_agent.partition("/")
            try:
                parsed_version = packaging.version.parse(raw_version)
            except Exception as caught_exception:
                invalid_version_message = (
                    f"Python Tiled client version is reported as {raw_version}. "
                    "This cannot be parsed as a valid version."
                )
                logger.warning(invalid_version_message)
                if isinstance(caught_exception, packaging.version.InvalidVersion):
                    warnings.warn(invalid_version_message)
            else:
                if (not parsed_version.is_devrelease) and (
                    parsed_version < MINIMUM_SUPPORTED_PYTHON_CLIENT_VERSION
                ):
                    return JSONResponse(
                        status_code=HTTP_400_BAD_REQUEST,
                        content={
                            "detail": (
                                f"Python Tiled client reports version {parsed_version}. "
                                f"Version {MINIMUM_SUPPORTED_PYTHON_CLIENT_VERSION} or higher "
                                "is needed to communicate with this Tiled server."
                            ),
                        },
                    )
        response = await call_next(request)
        return response

    @app.middleware("http")
    async def set_cookies(request: Request, call_next: RequestResponseEndpoint):
        "This enables dependencies to inject cookies that they want to be set."
        # Create some Request state, to be (possibly) populated by dependencies.
        request.state.cookies_to_set = []
        response = await call_next(request)
        for params in request.state.cookies_to_set:
            params.setdefault("httponly", True)
            params.setdefault("samesite", "lax")
            response.set_cookie(**params)
        return response

    app.openapi = partial(custom_openapi, app)
    app.dependency_overrides[get_settings] = override_get_settings

    @app.middleware("http")
    async def capture_metrics(request: Request, call_next: RequestResponseEndpoint):
        """
        Place metrics in Server-Timing header, in accordance with HTTP spec.
        """
        # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Server-Timing
        # https://w3c.github.io/server-timing/#the-server-timing-header-field
        # This information seems safe to share because the user can easily
        # estimate it based on request/response time, but if we add more detailed
        # information here we should keep in mind security concerns and perhaps
        # only include this for certain users.
        # Initialize a dict that routes and dependencies can stash metrics in.
        metrics = collections.defaultdict(lambda: collections.defaultdict(lambda: 0))
        request.state.metrics = metrics
        # Record the overall application time.
        with record_timing(metrics, "app"):
            response = await call_next(request)
        # Server-Timing specifies times should be in milliseconds.
        # Prometheus specifies times should be in seconds.
        # Therefore, we store as seconds and convert to ms for Server-Timing here.
        # That is what the factor of 1000 below is doing.
        response.headers["Server-Timing"] = ", ".join(
            f"{key};"
            + ";".join(
                (
                    f"{metric}={value * 1000:.1f}"
                    if metric == "dur"
                    else f"{metric}={value:.1f}"
                )
                for metric, value in metrics_.items()
            )
            for key, metrics_ in metrics.items()
        )
        return response

    metrics_config = server_settings.get("metrics", {})
    if metrics_config.get("prometheus", True):
        # PROMETHEUS_MULTIPROC_DIR puts prometheus_client in multiprocess mode
        # (for e.g. gunicorn) which uses a directory of memory-mapped files.
        # If that environment variable is set, check that the directory exists
        # and is writable.
        prometheus_multiproc_dir = os.getenv("PROMETHEUS_MULTIPROC_DIR", None)
        if prometheus_multiproc_dir:
            if not Path(prometheus_multiproc_dir).is_dir():
                raise ValueError(
                    "prometheus enabled and PROMETHEUS_MULTIPROC_DIR is set but "
                    f"({prometheus_multiproc_dir}) is not a directory"
                )
            if not os.access(prometheus_multiproc_dir, os.W_OK):
                raise ValueError(
                    "prometheus enabled and PROMETHEUS_MULTIPROC_DIR is set but "
                    f"({prometheus_multiproc_dir}) is not writable"
                )

        from . import metrics

        app.include_router(metrics.router, prefix="/api/v1")

        @app.middleware("http")
        async def capture_metrics_prometheus(
            request: Request, call_next: RequestResponseEndpoint
        ):
            try:
                response = await call_next(request)
            except Exception:
                # Make an ephemeral response solely for 'capture_request_metrics'.
                # It will only be used in the 'finally' clean-up block.
                only_for_metrics = Response(status_code=HTTP_500_INTERNAL_SERVER_ERROR)
                response = only_for_metrics
                # Now re-raise the exception so that the server can generate and
                # send an appropriate response to the client.
                raise
            finally:
                metrics.capture_request_metrics(request, response)

            # This is a *real* response (i.e., not the 'only_for_metrics' response).
            # An exception above would have triggered an early exit.
            return response

    @app.middleware("http")
    async def current_principal_logging_filter(
        request: Request, call_next: RequestResponseEndpoint
    ):
        # Set a default, which may be overridden below by authentication code.
        request.state.principal = None
        response = await call_next(request)
        current_principal.set(request.state.principal)
        return response

    app.add_middleware(
        CorrelationIdMiddleware,
        header_name="X-Tiled-Request-ID",
        generator=lambda: secrets.token_hex(8),
    )

    return app


def build_app_from_config(config, source_filepath=None, scalable=False):
    "Convenience function that calls build_app(...) given config as dict."
    kwargs = construct_build_app_kwargs(config, source_filepath=source_filepath)
    return build_app(scalable=scalable, **kwargs)


def app_factory():
    """
    Return an ASGI app instance.

    Use a configuration file at the path specified by the environment variable
    TILED_CONFIG or, if unset, at the default path "./config.yml".

    This is intended to be used for horizontal deployment (using gunicorn, for
    example) where only a module and instance or factory can be specified.
    """
    config_path = os.getenv("TILED_CONFIG", "config.yml")
    logger.info(f"Using configuration from {Path(config_path).absolute()}")

    parsed_config = parse_configs(config_path)

    # This config was already validated when it was parsed. Do not re-validate.
    web_app = build_app_from_config(parsed_config, config_path)
    uvicorn_config = parsed_config.get("uvicorn", {})
    print_server_info(
        web_app, host=uvicorn_config.get("host"), port=uvicorn_config.get("port")
    )
    return web_app


def __getattr__(name):
    """
    This supports tiled.server.app.app by creating app on demand.
    """
    if name == "app":
        try:
            return app_factory()
        except Exception as err:
            raise Exception("Failed to create app.") from err
    raise AttributeError(name)


def print_server_info(
    web_app: FastAPI,
    host: str = "127.0.0.1",
    port: int = 8000,
    include_api_key: bool = False,
):
    settings = get_settings()

    if settings.allow_anonymous_access:
        print(
            """
    Tiled server is running in "public" mode, permitting open, anonymous access
    for reading. Any data that is not specifically controlled with an access
    policy will be visible to anyone who can connect to this server.
""",
            file=sys.stderr,
        )
    if not web_app.state.authenticated and include_api_key:
        print(
            f"""
    Navigate a web browser or connect a Tiled client to:

    http://{host}:{port}?api_key={settings.single_user_api_key}

""",
            file=sys.stderr,
        )
    if settings.allow_anonymous_access:
        print(
            """    Because this server is public, the '?api_key=...' portion of
    the URL is needed only for _writing_ data (if applicable).

""",
            file=sys.stderr,
        )


class UnscalableConfig(Exception):
    pass


class TaskMap(TypedDict):
    background: list[AppTask]
    startup: list[AppTask]
    shutdown: list[AppTask]



import numpy
import xarray

from tiled.adapters.array import ArrayAdapter
from tiled.adapters.dataframe import TableAdapter
from tiled.adapters.mapping import MapAdapter
from tiled.adapters.xarray import DatasetAdapter

tree = MapAdapter(
    {
        "A": ArrayAdapter.from_array(numpy.ones((100, 100))),
        "B": ArrayAdapter.from_array(numpy.ones((100, 100, 100))),
        "C": TableAdapter.from_dict(
            {
                "x": 1 * numpy.ones(100),
                "y": 2 * numpy.ones(100),
                "z": 3 * numpy.ones(100),
            },
            npartitions=3,
        ),
        "D": DatasetAdapter.from_dataset(
            xarray.Dataset(
                data_vars={"temperature": ("time", [100, 99, 98])},
                coords={"time": [1, 2, 3]},
            )
        ),
    },
    metadata={"thing": "stuff"},
)



import inspect
import threading
import weakref
from typing import Callable, List

import anyio
import httpx
import msgpack
from starlette.testclient import TestClient
from websockets.sync.client import connect

from tiled.client.context import Context

Callback = Callable[["Subscription", dict], None]
"A Callback will be called with the Subscription calling it and a dict with the update."


class TestClientWebsocketWrapper:
    """Wrapper for TestClient websockets."""

    def __init__(self, http_client, uri: httpx.URL):
        self._http_client = http_client
        self._uri = uri
        self._websocket = None

    def connect(self, api_key: str):
        """Connect to the websocket."""
        query_string = self._uri.query.decode() if self._uri.query else ""
        path = self._uri.path + ("?" + query_string if query_string else "")
        self._websocket = self._http_client.websocket_connect(
            path, headers={"Authorization": f"Apikey {api_key}"}
        )
        self._websocket.__enter__()

    def recv(self, timeout=None):
        """Receive data from websocket with consistent interface."""
        return self._websocket.receive_bytes()

    def close(self):
        """Close websocket connection."""
        self._websocket.__exit__(None, None, None)


class RegularWebsocketWrapper:
    """Wrapper for regular websockets."""

    def __init__(self, http_client, uri: httpx.URL):
        self._http_client = http_client
        self._uri = uri
        self._websocket = None

    def connect(self, api_key: str):
        """Connect to the websocket."""
        self._websocket = connect(
            str(self._uri),
            additional_headers={"Authorization": f"Apikey {api_key}"},
        )

    def recv(self, timeout=None):
        """Receive data from websocket with consistent interface."""
        return self._websocket.recv(timeout=timeout)

    def close(self):
        """Close websocket connection."""
        self._websocket.close()


class Subscription:
    """
    Subscribe to streaming updates from a node.

    Parameters
    ----------
    context : tiled.client.Context
        Provides connection to Tiled server
    segments : list[str]
        Path to node of interest, given as a list of path segments
    start : int, optional
        By default, the stream begins from the most recent update. Use this parameter
        to replay from some earlier update. Use 1 to start from the first item, 0
        to start from as far back as available (which may be later than the first item),
        or any positive integer to start from a specific point in the sequence.
    """

    def __init__(self, context: Context, segments: List[str] = None, start: int = None):
        segments = segments or ["/"]
        self._context = context
        self._segments = segments
        params = {"envelope_format": "msgpack"}
        if start is not None:
            params["start"] = start
        scheme = "wss" if context.api_uri.scheme == "https" else "ws"
        path = "stream/single" + "/".join(f"/{segment}" for segment in segments)
        self._uri = httpx.URL(
            str(context.api_uri.copy_with(scheme=scheme)) + path,
            params=params,
        )
        name = f"tiled-subscription-{self._uri}"
        self._thread = threading.Thread(target=self._receive, daemon=True, name=name)
        self._callbacks = set()
        self._close_event = threading.Event()
        if isinstance(context.http_client, TestClient):
            self._websocket = TestClientWebsocketWrapper(context.http_client, self._uri)
        else:
            self._websocket = RegularWebsocketWrapper(context.http_client, self._uri)

    @property
    def context(self) -> Context:
        return self._context

    @property
    def segments(self) -> List[str]:
        return self._segments

    def add_callback(self, callback: Callback) -> None:
        """
        Register a callback to be run when the Subscription receives an update.

        The callback registry only holds a weak reference to the callback. If
        no hard references are held elsewhere in the program, the callback will
        be silently removed.

        Examples
        --------

        Simply subscribe the print function.

        >>> sub.add_callback(print)

        Subscribe a custom function.

        >>> def f(sub, data):
                ...

        >>> sub.add_callback(f)
        """

        def cleanup(ref: weakref.ref) -> None:
            # When an object is garbage collected, remove its entry
            # from the set of callbacks.
            self._callbacks.remove(ref)

        if inspect.ismethod(callback):
            # This holds the reference to the method until the object it is
            # bound to is garbage collected.
            ref = weakref.WeakMethod(callback, cleanup)
        else:
            ref = weakref.ref(callback, cleanup)
        self._callbacks.add(ref)

    def remove_callback(self, callback: Callback) -> None:
        """
        Unregister a callback.
        """
        self._callbacks.remove(callback)

    def _receive(self) -> None:
        "This method is executed on self._thread."
        TIMEOUT = 0.1  # seconds
        while not self._close_event.is_set():
            try:
                data_bytes = self._websocket.recv(timeout=TIMEOUT)
            except (TimeoutError, anyio.EndOfStream):
                continue
            data = msgpack.unpackb(data_bytes)
            for ref in self._callbacks:
                callback = ref()
                if callback is not None:
                    callback(self, data)

    def start(self) -> None:
        "Connect to the websocket and launch a thread to receive and process updates."
        if self._close_event.is_set():
            raise RuntimeError("Cannot be restarted once stopped.")
        API_KEY_LIFETIME = 30  # seconds
        needs_api_key = self.context.server_info.authentication.providers
        if needs_api_key:
            # Request a short-lived API key to use for authenticating the WS connection.
            key_info = self.context.create_api_key(
                expires_in=API_KEY_LIFETIME, note="websocket"
            )
            api_key = key_info["secret"]
        else:
            # Use single-user API key.
            api_key = self.context.api_key

        # Connect using the websocket wrapper
        self._websocket.connect(api_key)

        if needs_api_key:
            # The connection is made, so we no longer need the API key.
            # TODO: Implement single-use API keys so that revoking is not
            # necessary.
            self.context.revoke_api_key(key_info["first_eight"])
        self._thread.start()

    def stop(self) -> None:
        "Close the websocket connection."
        self._close_event.set()
        self._websocket.close()
        self._thread.join()
